----------------------------------------------------------------------------------------------------
----    ,   (   ). ----
----------------------------------------------------------------------------------------------------
module ArcExtract where

import Prelude hiding (catch)
import Control.Concurrent
import Control.Exception
import Control.Monad
import Data.Char
import Data.IORef
import Data.List
import Foreign.C.Types
import Foreign.Ptr
import Foreign.Marshal.Alloc
import Foreign.Marshal.Pool
import Foreign.Marshal.Utils
import System.Directory
import System.Environment
import System.IO

import Utils
import Files
import Process
import Errors
import FileInfo
import Compress
import Cmdline
import CUI
import ArhiveStructure

-- |   
runArchiveExtract command@Command{ cmd_arcname      = arcname
                                 , cmd_arc_filter   = arc_filter
                                 , opt_arc_basedir  = arc_basedir
                                 , opt_disk_basedir = disk_basedir
                                 , opt_arccmt_file  = arccmt_file
                                 } = do
    --  : find_archives -> buffer 10_000 -> read_dir -> buffer 10_000 -> arcExtract
  uiStartArchive command ""  --       
  bracket (archiveReadInfo command arc_basedir disk_basedir arc_filter arcname) (arcClose)$ \archive -> do
    uiPrintArcComment (arcComment archive)           --  
    --when (arccmt_file/="-" && arccmt_file/="--") $   --     ,   -z
    --  filePut arccmt_file (arcComment archive)       -- (      RAR)
    arcExtract command archive
  uiDoneArchive  --         

-- | 
arcExtract command arcinfo = do
  let process_file = case (cmd_name command) of    -- ,     
                       "t"  -> test_file
                       "e"  -> extract_file (fpBasename.fiDiskName) command
                       "x"  -> extract_file (fpFullname.fiDiskName) command
  --   UI          
  uiStartProcessing (map cfFileInfo (arcDirectory arcinfo))
  uiStartDirectory
  uiUnpackedBytes   (arcDirBytes  arcinfo)
  uiCompressedBytes (arcDirCBytes arcinfo)
  uiStartFiles
  --     
  decompress_pipe <- runAsyncP( decompress_PROCESS (uiCompressedBytes.i) )
  --        
  let (extract_files, skip_files)  =  partition isDecompressible (arcDirectory arcinfo)
  mapM_ (process_file decompress_pipe) extract_files     -- runP( enum_files |> decompress |> write_files )
  unless (null skip_files)$  do registerWarning$ SKIPPED_FAKE_FILES (length skip_files)
  --    
  sendP decompress_pipe Nothing
  joinP decompress_pipe

-- |    
test_file decompress_pipe compressed_file = do
  uiStartFile (cfFileInfo compressed_file)
  run_decompress decompress_pipe compressed_file (\buf size -> return ())
  return ()

-- |    
extract_file filename_func command decompress_pipe compressed_file = do
  let fileinfo  = cfFileInfo compressed_file
      filename  = filename_func fileinfo
  if (fiIsDir fileinfo)
    then do uiStartFile fileinfo
            createDirectoryHierarchy filename
    else do
  --   ,     
  whenM (can_be_extracted command filename fileinfo)$ do
    uiStartFile fileinfo
    buildPathTo filename
    outfile  <- fileCreate filename
    let closeOutfile ok = do   -- ,         ^Break
          fileClose outfile                              -- to do:   fileSetSize,           
          if ok || opt_keep_broken command
            then setFileDateTimeAttr filename fileinfo   --       ,   
            else fileRemove filename                     --  ,   
    do  --fileSetSize outfile (fiSize fileinfo)  --            
        ok <- run_decompress decompress_pipe compressed_file (fileWriteBuf outfile)
        closeOutfile ok
      `catch`
        \e -> do closeOutfile False; throwIO e

-- |   -      ?
--    1)   (-u/-f/-sync)
--                  2)     
--                  3) ,     -     
--                  4)   "-o"
--                  5)       
--
can_be_extracted cmd filename arcfile = do
  diskfile_exist <- fileExist filename
  if not diskfile_exist                         --      
    then return (opt_update_type cmd /= 'f')    --         ,  '-f'
    else do
  diskfile_time     <-  getFileDateTime filename
  let arcfile_newer  =  fiTime arcfile > diskfile_time   --    ,   ?
  let overwrite = case (opt_update_type cmd) of
                    'f' -> arcfile_newer
                    'u' -> arcfile_newer
                    's' -> error "--sync can't be used on extract"
                    'a' -> True
  case (overwrite, opt_overwrite cmd) of
    (False,  _ ) -> return False
    (True , "+") -> return True
    (True , "-") -> return False
    (True , "p") -> askOverwrite filename (ref_ask_overwrite cmd) arcfile_newer

-- |      CRC
run_decompress decompress_pipe compressed_file write_data = do
  crc <- ref aINIT_CRC                        --   CRC
  let writer buf len = do
        uiUnpackedBytes  (i len)              --     
        crc          .<- updateCRC buf len    --  CRC  
        write_data       buf len              --    
        send_backP       decompress_pipe ()   --    
  decompress_file decompress_pipe compressed_file writer
  acrc  <-  val crc >>== finishCRC            --    CRC
  when (cfCRC compressed_file /= acrc) $ do
    registerWarning$ BAD_CRC (fpFullname$ fiStoredName$ cfFileInfo compressed_file)
  return (cfCRC compressed_file == acrc)      --  True,   


-- |        
--        `writer`
decompress_file decompress_pipe compressed_file writer = do
  --    /     ,    0  -    ;)
  when (fiSize(cfFileInfo compressed_file) > 0  &&  isDecompressible compressed_file) $ do
    sendP decompress_pipe (Just compressed_file)
    repeat_whileM (receiveP decompress_pipe) ((>=0).snd) (uncurry writer)
    return ()

-- |  -  ?
isDecompressible = not . isFakeCompressor . blCompressor . cfArcBlock


-- |,    
decompress_PROCESS count_cbytes pipe = do
  cmd <- receiveP pipe
  case cmd of
    Nothing     -> return ()
    Just cfile' -> do
      cfile <- ref cfile'
      state <- ref (error "Decompression state is not initialized!")
      repeat_untilM $ do
        decompress_block cfile state count_cbytes pipe
        (x,_,_) <- val state
        return (x == -2)  --  ,      "Nothing"

-- |  -
decompress_block cfile state count_cbytes pipe = do
  cfile' <- val cfile
  let size         =  fiSize      (cfFileInfo cfile')
      pos          =  cfPos        cfile'
      block        =  cfArcBlock   cfile'
      compressor   =  blCompressor block
  state =: (0, pos, size)
  archiveBlockSeek block 0
  bytesLeft <- ref (blCompSize block)
  let reader buf size  =  do aBytesLeft   <-  val bytesLeft
                             let bytes     =  minI size aBytesLeft
                             len          <-  archiveBlockReadBuf block buf bytes
                             bytesLeft    -=  i len
                             count_cbytes     len
                             return len

  let writer (DataBuf buf len)  =  decompress_step cfile state pipe buf len
      writer  NoMoreData        =  return 0

  result <- ref 0   --  ,     writer
  let decompressP = de_compress_PROCESS freearcDecompress
      decompressa []         =  decompress_p aSTORING reader
      decompressa [a]        =  decompress_p a reader
      decompressa [a,b]      =  decompress_p b reader |> decompressP a
      decompressa [a,b,c]    =  decompress_p c reader |> decompressP b |> decompressP a
      decompressa [a,b,c,d]  =  decompress_p d reader |> decompressP c |> decompressP b |> decompressP a
      decompressa _          =  \pipe -> fail "More than 4 compressors in line currently is not supported"
  uiStartDeCompression   --    
  ; runFuncP (decompressa compressor) (undefined) (undefined) (writer .>>= writeIORef result) (val result)
  uiFinishDeCompression  --     (/            )

decompress_p comprMethod reader pipe = do
  freearcDecompress comprMethod reader (\buf size -> resend_data pipe (DataBuf buf size))
  resend_data pipe NoMoreData
  return ()

-- |,       /
data CompressionData = DataBuf (Ptr CChar) Int
                     | NoMoreData

-- |    /    
resend_data pipe x@DataBuf{}   =  sendP pipe x  >>  receive_backP pipe  --    ,   -
resend_data pipe x@NoMoreData  =  sendP pipe x  >>  return 0


-- |       
--     /
de_compress_PROCESS de_compress comprMethod pipe = do
  --    ,    ,      /
  remains <- ref$ Just (error "undefined remains:buf0", error "undefined remains:srcbuf", 0)
  let
    --  ""  
    read_data prevlen  --    
              dstbuf   -- ,     
              dstlen   --  
              = do     -- ->        0,   
      remains' <- val remains
      case remains' of
        Just (buf0, srcbuf, srclen)                   --    ,    
         | srclen>0  ->  copyData buf0 srcbuf srclen  --     /
         | otherwise ->  processNextInstruction       --    
        Nothing      ->  return prevlen               --  solid- ,   
      where
        --    srcbuf  dstbuf     
        copyData buf0 srcbuf srclen = do
          let len = srclen `min` dstlen    --  -     
          copyBytes dstbuf srcbuf len
          remains =: Just (buf0, srcbuf+:len, srclen-len)
          case () of
           _ | len==srclen -> do send_backP pipe (srcbuf-:buf0+srclen)               --   ,        /
                                 read_data (prevlen+len) (dstbuf+:len) (dstlen-len)  --   
             | len==dstlen -> return (prevlen+len)                                 --   
             | otherwise   -> read_data (prevlen+len) (dstbuf+:len) (dstlen-len)   --      

        --          
        processNextInstruction = do
          instr <- receiveP pipe
          case instr of
            DataBuf srcbuf srclen  ->  copyData srcbuf srcbuf srclen
            NoMoreData             ->  do remains =: Nothing;  return prevlen

  --    
  de_compress comprMethod (read_data 0) (\buf size -> resend_data pipe (DataBuf buf size))
  --   ,    
  resend_data pipe NoMoreData
  return ()


-- |     (writer  ).
--  (   state) :
--   1) block_pos -     
--   2) pos       - ,     (   )
--   3) size      -   (   )
-- ,       buf  len,  :
--   1)     ,    ( )
--   2)    ,     ( )
--   3)   -        ,
--               -      
--   4)     -      
--             
--   5)             
--          -       ,  decompress_block
--           ,   (     cfile)
--
decompress_step cfile state pipe buf len = do
  (block_pos, pos, size) <- val state
  if block_pos<0   -- ,     ,        
    then return (-1)   -- , ,   (.. PPMD ;) . : fail$ "Block not changed!!!"
    else do
  let skip_bytes = min (pos-block_pos) (i len)   --       
      data_start = buf +: skip_bytes             --  ,   
      data_size  = min size (i len-skip_bytes)   -- - ,   
  when (data_size>0) $ do    --     ,   
    sendP pipe (data_start, i data_size)  --        
    receive_backP pipe                    --   ,    
  state =: (block_pos+i len, pos+data_size, size-data_size)
  if data_size<size     --      
    then return len     --    
    else do             --       
  sendP pipe (error "End of decompressed data", -1)
  old_block  <-  cfArcBlock ==<< val cfile
  cmd <- receiveP pipe
  case cmd of
    Nothing -> do
      state =: (-2, error "undefined state.pos", error "undefined state.size")
      cfile =: error "undefined cfile"
      return (-1)

    Just cfile' -> do
      cfile =: cfile'
      let size   =  fiSize (cfFileInfo cfile')
          pos    =  cfPos      cfile'
          block  =  cfArcBlock cfile'
      if block/=old_block || pos<block_pos  --          ,  
        then do state =: (-1, error "undefined state.pos", error "undefined state.size")
                return (-1)   --  ,      
        else do state =: (block_pos, pos, size)            --    ,
                decompress_step cfile state pipe buf len   --      


{-# NOINLINE run_decompress #-}
{-# NOINLINE decompress_PROCESS #-}
{-# NOINLINE decompress_p #-}
{-# NOINLINE resend_data #-}
{-# NOINLINE de_compress_PROCESS #-}
