----------------------------------------------------------------------------------------------------
----     .                                                    ------
----      ,     :      ------
----   *                                                                          ------
----   *                                                                  ------
----   *    (recovery record,      ..)         ------
----     (  ,   ),        ------
----   ,                      ------
----     .                                                                   ------
----     :                                                      ------
----   *     (..     )          ------
----   * ,                                      ------
----   *                                         ------
----   *                                                          ------
----------------------------------------------------------------------------------------------------
module ArhiveStructure where

import Control.Monad
import Data.Array
import Data.HashTable as Hash
import Data.List
import Data.Word
import Foreign.Marshal.Pool
import Foreign.Ptr
import Foreign.Storable
import System.Directory

import GHC.PArr

import Utils
import Errors
import Files
import qualified ByteStream
import FileInfo
import CompressionLib
import Compress
import CUI

----------------------------------------------------------------------------------------------------
----     --------------------------------------------------------------
----------------------------------------------------------------------------------------------------

-- |     
data ArchiveInfo = ArchiveInfo {
                       arcArchive    :: Archive           --   
                     , arcDirBytes   :: !FileSize          --      
                     , arcDirCBytes  :: !FileSize          --      
                     , arcBlocks     :: ![ArchiveBlock]    --   
                     , arcDirectory  :: ![CompressedFile]  -- ,   
                     , arcLocked     :: !Bool              --    ?
                     , arcComment    :: !String            --   (to do:  !)
                     , arcSFXSize    :: !FileSize          --  SFX-,   
                     , arcPhantom    :: !Bool              -- True,       (  main_archive)
                     }

-- |  ,      
arcClose arc  =  unless (arcPhantom arc) $  do archiveClose (arcArchive arc)

-- |,  ,      
-- (  ,   )
phantomArc  =  ArchiveInfo (error "phantom arcArchive") 0 0 [] [] False "" 0 True

-- | archiveReadInfo,    ,   `arcname`   
phantomArchiveReadInfo command arc_basedir disk_basedir filter_f arcname = do
  exist <- fileExist arcname
  if exist
    then archiveReadInfo command arc_basedir disk_basedir filter_f arcname
    else return phantomArc

-- |        ,       
aSCAN_MAX = 256

-- |    (   )
archiveReadInfo command               --      
                arc_basedir           --     (""   )
                disk_basedir          --     (""   /)
                filter_f              --       
                arcname = do          --  ,  
  archive <- archiveOpen arcname
  arcsize <- archiveGetSize archive
  let scan_bytes = min aSCAN_MAX arcsize  --  256    ,    :)

  withPool $ \pool -> do
    --  `scan_bytes`          ,
    --     
    buf         <- archiveMallocReadBuf pool archive (arcsize-scan_bytes) (i scan_bytes)
    footer_pos  <- archiveFindBlockDescriptor buf (i scan_bytes)

    --    
    footer_descriptor <-
      archiveReadBlockDescriptor
        archive                                         --  
        arcname                                         --   -    
        (arcsize - scan_bytes + i footer_pos)   --      
        (buf `plusPtr` footer_pos)                      --     
        (i scan_bytes - footer_pos)           --   

    --    (FOOTER_BLOCK)       
    (control_blocks, arcLocked, arcComment)  <-  archiveReadFooterBlock arcname footer_descriptor

    --    ,   FOOTER_BLOCK
    let dir_blocks  =  filter ((DIR_BLOCK==).blType) control_blocks
    files  <-  (`concatMapM` dir_blocks) $ \block -> do
      withPool $ \pool -> do
        buf <- archiveBlockReadAll pool block
        archiveReadDir arc_basedir disk_basedir archive (blPos block) filter_f (return (buf, i$ blOrigSize block))

    --     ArchiveInfo     
    return ArchiveInfo
             { arcArchive   = archive
             , arcDirBytes  = sum$ map blOrigSize dir_blocks
             , arcDirCBytes = sum$ map blCompSize dir_blocks
             , arcBlocks    = control_blocks
             , arcDirectory = files
             , arcLocked    = arcLocked
             , arcComment   = arcComment
             , arcSFXSize   = minimum (map blPos control_blocks)   --  SFX =    Header 
             , arcPhantom   = False
             }

{-# NOINLINE archiveReadInfo #-}


----------------------------------------------------------------------------------------------------
---- //     ---------------------------------------
----------------------------------------------------------------------------------------------------

aSIGNATURE = make4byte 65 114 67 1 :: Word32   --      

-- |      `block`  CRC  
archiveWriteBlockDescriptor block receiveBuf sendBuf = do
  crc  <- ref aINIT_CRC
  let sendBuf_UpdatingCRC buf size len  =  do crc .<- updateCRC buf len;  sendBuf buf size len
  ByteStream.writeAll receiveBuf sendBuf_UpdatingCRC (return ())
    --         -    ,   :
    --   ,  ,  ,      , CRC  
    (aSIGNATURE, blType block, blCompressor block, blOrigSize block, blCompSize block, blCRC block)
  acrc <- val crc
  --    CRC  
  ByteStream.writeAll receiveBuf sendBuf (return ())
    (finishCRC acrc)

-- |           ,
--         `arcpos`
archiveReadBlockDescriptor archive arcname arcpos buf bufsize = do
  --   CRC  
  right_crc  <- peekByteOff buf (bufsize - sizeOf (undefined::CRC))
  descriptor_crc <- calcCRC buf (bufsize - sizeOf (undefined::CRC))
  when (descriptor_crc/=right_crc) $ do
    registerError$ BROKEN_ARCHIVE ("block descriptor at pos "++ show arcpos ++ " is corrupted")
  --   :
  -- ,  ,  ,      , CRC  
  (sign, block_type, compressor, origsize, compsize, crc)  <-  ByteStream.readMemory buf bufsize
  let pos   =  blDecodePosRelativeTo arcpos compsize  --     
      block =  ArchiveBlock archive block_type compressor pos origsize compsize crc undefined
  when (sign/=aSIGNATURE || pos<0) $ do
    registerError$ BROKEN_ARCHIVE (block_name block ++ " is corrupted")
  return block

-- |     
archiveFindBlockDescriptor buf size  =  go (size - sizeOf aSIGNATURE)
  where go pos | pos<0     = registerError$ BROKEN_ARCHIVE errormsg
               | otherwise = do x <- peekByteOff buf pos
                                if (x==aSIGNATURE)
                                  then return pos
                                  else go (pos-1)
        errormsg = "archive signature was not found"
                     ++ if size == aSCAN_MAX
                          then " in last "++show size++" bytes of archive"
                          else " in archive"

{-# NOINLINE archiveWriteBlockDescriptor #-}
{-# NOINLINE archiveReadBlockDescriptor #-}
{-# NOINLINE archiveFindBlockDescriptor #-}


----------------------------------------------------------------------------------------------------
----     (HEADER_BLOCK) -------------------------------------------------
----------------------------------------------------------------------------------------------------

-- |,      -      
aARCHIVE_SIGNATURE = (aSIGNATURE, make4byte 0 0 2 5)

-- |     (HEADER_BLOCK)  
archiveWriteHeaderBlock receiveBuf sendBuf = do
  ByteStream.writeAll receiveBuf sendBuf (return ()) $
    aARCHIVE_SIGNATURE


----------------------------------------------------------------------------------------------------
----       (FOOTER_BLOCK) ------------------------------------------
----------------------------------------------------------------------------------------------------

-- |    (FOOTER_BLOCK)
archiveWriteFooterBlock control_blocks arcLocked arcComment arcpos receiveBuf sendBuf = do
  stream <- ByteStream.create receiveBuf sendBuf (return ())
  ByteStream.write        stream (map (blockToTuple arcpos) control_blocks)   --    ,
  ByteStream.write        stream arcLocked                                    -- ...     
  ByteStream.writeInteger stream (length arcComment)                          -- ...    (  ,              )
  ByteStream.writeList    stream arcComment
  ByteStream.closeOut     stream

-- |      (FOOTER_BLOCK)
archiveReadFooterBlock arcname block @ ArchiveBlock {
                                           blArchive     = archive
                                         , blType        = block_type
                                         , blPos         = pos
                                         , blOrigSize    = origsize
                                       } = do
  when (block_type/=FOOTER_BLOCK) $
    registerError$ BROKEN_ARCHIVE "footer block not found"
  withPool $ \pool -> do   --   ,       
    buf <- archiveBlockReadAll pool block         --      
    stream <- ByteStream.openMemory buf (i origsize)
    control_blocks <- ByteStream.read stream      --    ,
    arcLocked      <- ByteStream.read stream      -- ...     
    arcComment     <- ByteStream.readInteger stream >>= ByteStream.readList stream  -- ...    (  ,              )
    ByteStream.closeIn stream
    return (map (tupleToBlock archive pos) control_blocks, arcLocked, arcComment)    --   ArchiveBlock   

{-# NOINLINE archiveWriteFooterBlock #-}
{-# NOINLINE archiveReadFooterBlock #-}


----------------------------------------------------------------------------------------------------
----   ( ,   ) ----------------------------------------------
----------------------------------------------------------------------------------------------------

-- | 
data ArchiveBlock = ArchiveBlock {
                        blArchive     :: !Archive      -- ,     
                      , blType        :: !BlockType    --  
                      , blCompressor  :: !Compressor   --  
                      , blPos         :: !FileSize     --     
                      , blOrigSize    :: !FileSize     --     
                      , blCompSize    :: !FileSize     --     
                      , blCRC         :: !CRC          -- CRC   (   )
                      , blFiles       ::  Int          --    solid-
                    } deriving (Eq)

-- |       .     
--  `arcpos` -     ,     
blockToTuple              arcpos (ArchiveBlock _ t c p o s crc f) = (t,c,arcpos-p,o,s,crc)
tupleToBlock      archive arcpos (t,c,p,o,s,crc)    = (ArchiveBlock archive t c (arcpos-p) o s crc undefined)
tupleToSolidBlock archive arcpos (t,c,p,o,s,crc,f)  = (ArchiveBlock archive t c (arcpos-p) o s crc f)

--    ()       
blEncodePosRelativeTo arcpos arcblock  =  arcpos - blPos arcblock
blDecodePosRelativeTo arcpos offset    =  arcpos - offset

-- | 
block_name block  =  (case (blType block) of
                        DESCR_BLOCK  -> "block descriptor"
                        HEADER_BLOCK -> "header block"
                        DATA_BLOCK   -> "data block"
                        DIR_BLOCK    -> "directory block"
                        FOOTER_BLOCK -> "footer block"
                        -- _         -> "block of unknown type"
                      ) ++ " at pos "++ show (blPos block)

-- |   (    ,     !)
data BlockType = DESCR_BLOCK       -- |     (    ,   ,  DATABLOCK)
               | HEADER_BLOCK      -- |      (    )
               | DATA_BLOCK        -- |  
               | DIR_BLOCK         -- |  
               | FOOTER_BLOCK      -- |       (    )
     deriving (Eq,Enum)

instance ByteStream.BufferData BlockType  where
  write buf a = ByteStream.writeInteger buf (fromEnum a)
  read  buf   = ByteStream.readInteger  buf >>== toEnum

--    
archiveBlockSeek    block pos       =  archiveSeek    (blArchive block) (blPos block + pos)
archiveBlockRead    block size      =  archiveRead    (blArchive block) size
archiveBlockReadBuf block buf size  =  archiveReadBuf (blArchive block) buf size

-- | ,        CRC
archiveBlockReadAll pool block @ ArchiveBlock {
                                     blArchive     = archive
                                   , blType        = block_type
                                   , blCompressor  = compressor
                                   , blPos         = pos
                                   , blCRC         = right_crc
                                 } = do
  let origsize = i$ blOrigSize block
      compsize = i$ blCompSize block
  withPool $ \pool2 -> do
    compbuf <- archiveMallocReadBuf pool2 archive pos compsize
    origbuf <- pooledMallocBytes pool (origsize+8)  -- +8 - -   ByteStream :(
    decompressed_size <- decompressMem (encode_compressor compressor) compbuf compsize origbuf origsize
    crc <- calcCRC origbuf origsize
    when (crc/=right_crc || decompressed_size/=origsize) $ do
      registerError$ BROKEN_ARCHIVE (block_name block ++ " is corrupted")
    return origbuf

-- |        
archiveMallocReadBuf pool archive pos size = do
  buf         <- pooledMallocBytes pool (size+8)  -- +8 - -   ByteStream :(
  archiveSeek    archive pos
  archiveReadBuf archive buf size
  return buf



----------------------------------------------------------------------------------------------------
----    -------------------------------------------------------------------------
----------------------------------------------------------------------------------------------------

-- |  
aTAG_END = 0::Integer   --    

-- | `dirdata`          `sendBuf`
archiveWriteDir dirdata     --   (block :: ArchiveBlock, directory :: [CompressedFile])
                arcpos      --   ,    
                receiveBuf  -- "(buf,size) <- receiveBuf"       `size`
                sendBuf     -- "sendBuf buf size len"       
                = do
  statMark "Writing directory"
  let blocks    = map fst dirdata          :: [ArchiveBlock]    --  -,    
      cfilelist = concatMap snd dirdata    :: [CompressedFile]  --    -   ,      !
      filelist  = map cfFileInfo cfilelist :: [FileInfo]        --    

  -- 0. C  ,        `receiveBuf`  `sendBuf`
  stream <- ByteStream.create receiveBuf sendBuf (return ())
  let write          =  ByteStream.write          stream   -- shortcuts     
      writeLength    =  ByteStream.writeInteger   stream . length
      writeList      =  ByteStream.writeList      stream
      writeIntegers  =  mapM_ (ByteStream.writeInteger stream)
      writeTagged     tag x   =  write tag >> write x     --    -   
      writeTaggedList tag xs  =  write tag >> writeList xs

  -- 1.      -     
  writeLength dirdata               -- - .       :
  mapM_ (writeLength.snd) dirdata                        -- - 
  writeList$ map (blCompressor                ) blocks   --  
  writeList$ map (blEncodePosRelativeTo arcpos) blocks   --      
  writeList$ map (blCompSize                  ) blocks   --     

  -- 2.      
    --       ,    filelist
  (n, dirnames, dir_numbers)  <-  enumDirectories filelist
  statMark$ "Found "++show n++" directory names"
  writeLength dirnames  -- ,     Compressor==[String]
  writeList   dirnames

  -- 3.       CompressedFile/FileInfo
    -- to do:  RLE- ?
    -- to do:  Unicode-.       getDirectoryContents :(
  writeList$ map (fpBasename.fiStoredName)  filelist     --  
  writeIntegers                             dir_numbers  --  
  writeList$ map fiSize                     filelist     --  
  writeList$ map fiTime                     filelist     --  
  writeList$ map fiIsDir                    filelist     --  
  -- cfArcBlock  cfPos  ,      
  writeList$ map cfCRC                      cfilelist    -- CRC

  -- 4.  ,   ,   -    
  write aTAG_END  --    ,        

  -- 5. ! :)
  ByteStream.closeOut stream
  --statMark "Directory written"


--     -   +       
enumDirectories filelist = do
  --   Stored           - `table`.
  --   ,     -   ,
  --    -     -    , 
  --    `counter`,       `dirnames`.
  --  , - `table`      
  --      `dirnames`.
  table    <- Hash.new (==) fpHash               --     
  counter  <- ref (0::Int)                       --     
  dirnames <- newList                            --   

  --       `dirnames`    ,
  --     `filelist`
  dir_numbers <- (`mapM` filelist) $ \fileinfo -> do
    let storedName  =  fiStoredName fileinfo    -- ,     
        dirname     =  fpParent storedName      -- ,    
    x <- Hash.lookup table dirname              --       ?
    case x of                                   --  , 
      Nothing -> do n <- val counter               --      
                    Hash.insert table dirname n    --     
                    counter   += 1                  --   
                    dirnames <<= fpDirectory storedName  --       
                    return n                       --   
      Just n  -> return n                          --  -    

  n         <- val counter
  dirnames' <- listVal dirnames
  return (n, dirnames', dir_numbers)



----------------------------------------------------------------------------------------------------
----    -------------------------------------------------------------------------
----------------------------------------------------------------------------------------------------

-- | ,   `archiveWriteDir`
archiveReadDir arc_basedir   --    
               disk_basedir  --    
               archive       --  
               arcpos        --   ,    
               filter_f      --   
               receiveBuf    -- "(buf,size) <- receiveBuf"       `size`
               = do
  statMark "Decoding directory"

  -- 0. C  ,        `receiveBuf`
  stream <- ByteStream.open receiveBuf (\a b c->return ()) (return ())
  let read           = ByteStream.read stream           -- shortcuts     
      readList       = ByteStream.readList stream
      readInteger    = ByteStream.readInteger stream
      readLength     = readInteger
      readIntegers n = replicateM n readInteger

  -- 1.    
  num_of_blocks <- readLength                   -- - 
  --    :
  num_of_files  <- readIntegers num_of_blocks   -- - 
  blCompressors <- readList     num_of_blocks   --  
  blOffsets     <- readList     num_of_blocks   --      
  blCompSizes   <- readList     num_of_blocks   --     

  -- 2.   
  --         
  total_dirs  <-  readLength
  --   
  storedName  <-  readList total_dirs >>== toP

  -- 3.        CompressedFile/FileInfo
  let total_files = sum num_of_files         --  -   
  names        <- readList     total_files   --   (  )
  dir_numbers  <- readIntegers total_files   --      
  sizes        <- readList     total_files   --  
  times        <- readList     total_files   --   
  dir_flags    <- readList     total_files   --   " ?"
  crcs         <- readList     total_files   -- CRC 

  -- 4.  ,   ,   -    
  repeat_whileM (read) (/=aTAG_END) $ \tag -> do
    -- to do:      -  ByteStream.skipBytes stream =<< readInteger
    registerError$ BROKEN_ARCHIVE ("can't skip optional field TAG="++show tag++" in archive directory")
    return ()

  -- 5. ! :)
  ByteStream.closeIn stream
  statMark "Directory decoded"


  ------------------------------------------------------------------------------------------------
  --       -----------------------------------------------
  ------------------------------------------------------------------------------------------------
  let -- ,      Filtered/Disk name (  Stored name    )
      drop_arc_basedir  = if arc_basedir>""  then drop (length arc_basedir + 1)  else id
      filteredName      = fmap (drop_arc_basedir)               storedName
      diskName          = fmap (disk_basedir `joinDirFilename`) filteredName
      -- ,      PackedFilePath
      storedInfo        = fmap packParentDirPath storedName
      filteredInfo      = fmap packParentDirPath filteredName
      diskInfo          = fmap packParentDirPath diskName
      --    -  :        ("-ap")
      dirIncludedArray  = fmap (arc_basedir `isParentDirOf`) storedName
      dirIncluded       = if arc_basedir==""  then const True  else (dirIncludedArray!:)

  --   Maybe FileInfo (Nothing   ,   
  --   ("-ap"),       )
  let make_fi dir name size time dir_flag =
        if dirIncluded dir && filter_f fileinfo  then Just fileinfo  else Nothing

        where fileinfo = FileInfo { fiFilteredName  =  if arc_basedir>""  then fiFilteredName  else fiStoredName
                                  , fiDiskName      =  if arc_basedir>"" || disk_basedir>""  then fiDiskName  else fiStoredName
                                  , fiStoredName    =  fiStoredName
                                  , fiSize          =  size
                                  , fiTime          =  time
                                  , fiIsDir         =  dir_flag
                                  }
              fiStoredName    =  packFilePath2 (storedInfo  !:dir) (storedName  !:dir) name
              fiFilteredName  =  packFilePath2 (filteredInfo!:dir) (filteredName!:dir) name
              fiDiskName      =  packFilePath2 (diskInfo    !:dir) (diskName    !:dir) name

  --   FileInfo   ,   
  let fileinfos = zipWith5 make_fi dir_numbers names sizes times dir_flags

  --    .
  --       ,    .
  --           
  let filesizes = splitByLens num_of_files sizes
  let blocks = map (tupleToSolidBlock archive arcpos) $
               zip7 (repeat DATA_BLOCK)
                    blCompressors
                    blOffsets
                    (map sum filesizes)
                    blCompSizes
                    (repeat 0)
                    num_of_files

  --      ,      :)
  let arcblocks = concat [ replicate files_in_block blockDescriptor
                           | files_in_block  <- num_of_files  -- -     
                           | blockDescriptor <- blocks        --   
                         ]

  --            .
  -- filesizes -    ,    .
  --  ,        ,   
  -- " ".  [0]     ,
  --     ,     :)
  --  ,   num_of_files = [1..4]
  --                    sizes = [1..10]
  --                 filesizes = [[1],[2,3],[4,5,6],[7,8, 9,10]]
  --                  positions = [ 0,  0,2,  0,4,9,  0,7,15,24]
  let positions = concatMap ((0:) . init . scanl1 (+)) filesizes

  --          ,    
  return [ CompressedFile fileinfo arcblock pos crc
           | (Just fileinfo, arcblock, pos, crc)  <-  zip4 fileinfos arcblocks positions crcs
         ]



----------------------------------------------------------------------------------------------------
----   (  ,     ) -----------------------------
----------------------------------------------------------------------------------------------------

-- |File to compress: either file on disk or compressed file in existing archive
data FileToCompress = DiskFile {
                          cfFileInfo :: FileInfo
                      }
                    | CompressedFile {
                          cfFileInfo :: FileInfo
                        , cfArcBlock :: ArchiveBlock    -- Archive datablock which contains file data
                        , cfPos      :: FileSize        -- Starting byte of file data in datablock
                        , cfCRC      :: CRC             -- File's CRC
                      }

-- |Assign type synonym because variant label can't be used in another type declarations
type CompressedFile = FileToCompress

-- |  ,   ArcBlock/Pos   
makeCompressedFile crc fi =
  CompressedFile { cfFileInfo = fi
                 , cfArcBlock = error "undefined cfArcBlock"
                 , cfPos      = error "undefined cfPos"
                 , cfCRC      = crc
                 }

-- | ,    -    ,    
isCompressedFile CompressedFile{} = True
isCompressedFile DiskFile{}       = False

-- |,     -     solid-
isWholeSolidBlock files @ (CompressedFile {cfArcBlock=block}:_) =
  blFiles block == length files       &&    --      ,   -,       ,
  all        isCompressedFile files   &&    --       ,
  isEqOn     cfArcBlock       files   &&    --     
  isSortedOn cfPos            files         --        -

isWholeSolidBlock _ = False

-- |    ,     - (,   -)
solidBlockLen [] = 0
solidBlockLen files = let n = blFiles (cfArcBlock (head files)) in
                      if isCompressedFile (head files)           --      
                         && isWholeSolidBlock (take n files)     --      -
                      then n
                      else 0
