----------------------------------------------------------------------------------------------------
----       (CUI - Console User Interface).  ------
----------------------------------------------------------------------------------------------------
module CUI where

import Control.Monad
import Data.IORef
import Data.List
import Foreign
import Foreign.C
import Numeric
import System.CPUTime
import System.Directory
import System.Exit
import System.IO
import System.IO.Unsafe
import System.Time

import Utils
import Errors
import Files
import FileInfo
import Cmdline
import Compress

-- |         ,   
--      
data UI_State = UI_State {
    command         :: !Command     --   
  , total_files     :: !Integer     -- - ,    
  , total_bytes     :: !FileSize    --     (  )
  , indicator       :: !Indicator   --    (, ,  )
  , title_state     :: !(Maybe ActionState)   --       
  , indicator_state :: !(Maybe ActionState)   --       
  , datatype        ::  DataType    --      : // 
  --    ,     ,  
  --    :
  ,    files        :: !Integer     -- -   
  ,    bytes        :: !FileSize    --       
  ,    cbytes       :: !FileSize    --       
  --     (   ):
  ,    dirs         :: !Integer     -- -      
  ,    dir_bytes    :: !FileSize    --       
  ,    dir_cbytes   :: !FileSize    --       
  --  ,  ,      -      (      ):
  ,    fake_bytes   :: !FileSize    --       
  ,    fake_cbytes  :: !FileSize    --       
  }

-- |     : // 
data DataType = File | Dir | CData   deriving (Eq)

ref_command               =  unsafePerformIO$ newIORef$ error "undefined ref_command"
refStartArchiveTime       =  unsafePerformIO$ newIORef$ error "undefined refStartArchiveTime"
refArchiveProcessingTime  =  unsafePerformIO$ newIORef$ error "undefined refArchiveProcessingTime"
ref_ui_state              =  unsafePerformIO$ newIORef$ error "undefined ref_ui_state"
ref_first_time            =  unsafePerformIO$ newIORef$ error "undefined ref_first_time"

-- |   
uiStartProgram = do
  writeIORef ref_first_time True
  --putStrLn$ "ARC 0.24 by Bulat Ziganshin bulat_z##mail.ru  2.05.2005"
  putStr$ "ARC 9.12 "

-- |   
uiStartCommand command = do
  writeIORef ref_command command

-- |    
uiStartArchive command method = do
  writeIORef refStartArchiveTime =<< getClockTime
  let cmd = cmd_name command
      arcname = cmd_arcname command
  writeIORef ref_command command

  first_time  <-  readIORef ref_first_time
  when (not first_time) $  putStr "\n\n"
  writeIORef ref_first_time False

  exist <- fileExist arcname
  putStrLn$ (msgStart cmd exist) ++ (str2terminal arcname) ++ method

-- |     
uiStartProcessing filelist = do
  writeIORef refArchiveProcessingTime 0
  command  <-  readIORef ref_command
  let cmd = cmd_name command
      total_files = i$ length filelist
      total_bytes = sum (map fiSize filelist)
      ui_state = UI_State {
          command         = command
        , total_files     = total_files
        , total_bytes     = total_bytes
        , indicator       = select_indicator command total_bytes
        , title_state     = Nothing
        , indicator_state = Nothing
        , datatype        = error "internal CUI error"
        , files           = 0
        , bytes           = 0
        , cbytes          = 0
        , dirs            = 0
        , dir_bytes       = 0
        , dir_cbytes      = 0
        , fake_bytes      = 0
        , fake_cbytes     = 0
        }
  writeIORef ref_ui_state ui_state
  putStr$ msgDo cmd ++ show_files3 total_files ++ ", "
  putStr$ ""        ++ show3 total_bytes ++ " bytes"
  when (opt_indicator command == "1") $ do
    putStr$ ". Processed " ++ percents ui_state
  hFlush stdout

-- |  / 
uiStartFiles = do
  modifyIORef ref_ui_state $ \ui_state ->
    ui_state { datatype = File }

-- |  /  
uiStartDirectory = do
  modifyIORef ref_ui_state $ \ui_state ->
    ui_state { datatype = Dir
             , dirs     = dirs ui_state + 1 }

-- |  /   
uiStartControlData = do
  modifyIORef ref_ui_state $ \ui_state ->
    ui_state { datatype = CData }

-- |  / 
uiStartFile fileinfo = do
  modifyIORefIO ref_ui_state $ \ui_state -> do
    let cmd           =  command ui_state
        filename      =  str2terminal$ fpFullname$ fiStoredName$ fileinfo
        printed_name  =  if fiIsDir fileinfo  then (filename++"/")  else filename
    when (opt_indicator cmd == "2") $ do
      putStrLn$  if (files ui_state > 0) then (clear_percents ui_state) else ""
      putStr$    left_justify 72 ((msgFile$ cmd_name cmd) ++ printed_name)
      putStr$    " " ++ percents ui_state
      hFlush     stdout
    return ui_state {
        datatype    = File
      , files       = files ui_state + 1
      }

-- | total_bytes  ui_state
uiCorrectTotal addMe = do
  when (addMe/=0) $ do
    modifyIORef ref_ui_state $ \ui_state ->
      ui_state { total_bytes = total_bytes ui_state + addMe }

-- |      
uiFakeFiles filelist compsize = do
  let origsize  =  sum (map fiSize filelist)
  modifyIORef ref_ui_state $ \ui_state ->
    ui_state { datatype    = File
             , files       = (files ui_state) + (i$ length filelist)
             , fake_bytes  = (fake_bytes  ui_state) + origsize
             , fake_cbytes = (fake_cbytes ui_state) + compsize
             }
  uiUnpackedBytes   origsize
  uiCompressedBytes compsize

-- |,    -    (  
--       ,           )
uiUnpackedBytes len = do
  --   :     - 
  -- .    ,         
  --   .       ,   
  --    .        :)
  uiUpdateProgressIndicator
  modifyIORef ref_ui_state $ \ui_state ->
    case (datatype ui_state) of
      File  ->  ui_state {     bytes =     bytes ui_state + len }
      Dir   ->  ui_state { dir_bytes = dir_bytes ui_state + len }
      CData ->  ui_state

-- |,    -    (, 
--   ,        , 
--       - )
uiCompressedBytes len = do
  modifyIORef ref_ui_state $ \ui_state ->
    case (datatype ui_state) of
      File  ->  ui_state {     cbytes =     cbytes ui_state + len }
      Dir   ->  ui_state { dir_cbytes = dir_cbytes ui_state + len }
      CData ->  ui_state

-- |   [   ]   (    )
uiUpdateProgressIndicator = do
  ui_state <- val ref_ui_state
  aProgramTerminated <- val programTerminated
  when (indicator ui_state /= NoIndicator  &&  datatype ui_state == File  &&  not aProgramTerminated) $ do
    let p          =  percents ui_state
        cmd        =  cmd_name$ command ui_state
        direction  =  if (cmdType cmd == ADD_CMD)  then " => "  else " <= "
        arcname    =  str2terminal$ cmd_arcname$ command ui_state
    current_time <- getClockTime
    --     ,    0.5 
    new_title_state     <- do_if (title_state ui_state) current_time 0.5
                                 setConsoleTitle
                                 (dropWhile (' '==) p ++ direction ++ arcname)
    --     ,    0.05 
    new_indicator_state <- do_if (indicator_state ui_state) current_time 0.05
                                 (putStr .>> hFlush stdout)
                                 (back_percents ui_state ++ p)
    writeIORef ref_ui_state ui_state { title_state     = new_title_state
                                     , indicator_state = new_indicator_state
                                     }

-- |       
uiStartDeCompression = do
  time <- getCPUTime
  modifyIORef refArchiveProcessingTime (\pt -> pt-time)

-- |       
uiFinishDeCompression = do
  time <- getCPUTime
  modifyIORef refArchiveProcessingTime (\pt -> pt+time)

-- |    ->       
uiDoneArchive = do
  ui_state @ UI_State { command       = command
                      , total_files   = total_files
                      , total_bytes   = total_bytes
                      , files         = files
                      , bytes         = bytes
                      , cbytes        = cbytes
                      , dirs          = dirs
                      , dir_bytes     = dir_bytes
                      , dir_cbytes    = dir_cbytes
                      , fake_bytes    = fake_bytes
                      , fake_cbytes   = fake_cbytes }  <-  readIORef ref_ui_state
  let cmd = cmd_name command
  if (opt_indicator command == "2")
    then putStrLn$  if (files > 0) then (clear_percents ui_state) else ""
    else putStr$    "\r"

  --  
  putStr$ msgDone cmd    ++ show_files3 files ++ ", "
  putStr$ ""             ++ show_ratio cmd bytes cbytes
  when (dir_bytes>10*1000) $ do   --         - 
    putStrLn$ ""
    putStr$ "Directory "   ++ if (dirs>1) then ("have " ++ show3 dirs ++ " chunks, ") else ""
    putStr$ ""             ++ show_ratio cmd dir_bytes dir_cbytes

  --       /
  processingTime <- readIORef refArchiveProcessingTime
  let secs   = i processingTime / 1e12   -- ,    /
  real_secs <- return_real_secs          --       
  putStrLn$ ""
  putStr$   msgStat cmd ++ "time "   ++ showFFloat (Just 2) secs " secs"
  putStr$   ", speed "               ++ show3(round$ i (bytes-fake_bytes)/1000/(max secs 1e-6)) ++ " kb/s"
  putStr$   ". Total "               ++ showFFloat (Just 2) real_secs " secs"

  hFlush stdout
  resetConsoleTitle
  return (1,files,bytes,cbytes)

-- |  ,       
uiDoneCommand Command{cmd_name=cmd} totals = do
  let sum4 (a0,b0,c0,d0) (a,b,c,d)   =  (a0+a,b0+b,c0+c,d0+d)
      (counts, files, bytes, cbytes) =  foldl1 sum4 totals
  when (counts > 1)$ do
    putStr$ "\n\nTotal: "++show3 counts++" archives, "
                         ++show3 files ++" files, "
    if (cbytes>=0)
      then putStr$ show_ratio cmd bytes cbytes
      else putStr$ show3 bytes++" bytes"
    hFlush stdout

-- |  
uiDoneProgram = resetConsoleTitle

{-# NOINLINE uiStartArchive #-}
{-# NOINLINE uiStartProcessing #-}
{-# NOINLINE uiStartFile #-}
{-# NOINLINE uiCorrectTotal #-}
{-# NOINLINE uiUnpackedBytes #-}
{-# NOINLINE uiCompressedBytes #-}
{-# NOINLINE uiUpdateProgressIndicator #-}
{-# NOINLINE uiDoneArchive #-}
{-# NOINLINE uiDoneCommand #-}


----------------------------------------------------------------------------------------------------
----  ,    ------------------------------------------
----------------------------------------------------------------------------------------------------

msgStart cmd arcExist =
                case (cmdType cmd, arcExist) of
                  (ADD_CMD,     False)  ->  "Creating archive: "
                  (ADD_CMD,     True)   ->  "Updating archive: "
                  (LIST_CMD,    _)      ->  "Listing archive: "
                  (TEST_CMD,    _)      ->  "Testing archive: "
                  (EXTRACT_CMD, _)      ->  "Extracting archive: "

msgDo cmd    =  case (cmdType cmd) of
                  ADD_CMD     -> "Compressing "
                  TEST_CMD    -> "Testing "
                  EXTRACT_CMD -> "Extracting "

msgFile      =  ("  "++).msgDo

msgDone cmd  =  case (cmdType cmd) of
                  ADD_CMD     -> "Compressed "
                  TEST_CMD    -> "Tested "
                  EXTRACT_CMD -> "Extracted "

msgStat cmd  =  case (cmdType cmd) of
                  ADD_CMD     -> "Compression "
                  TEST_CMD    -> "Testing "
                  EXTRACT_CMD -> "Extraction "

-- | "file"  "files",    -
show_files3 1 = "1 file"
show_files3 n = show3 n ++ " files"



----------------------------------------------------------------------------------------------------
----   ---------------------------------------------------------------------------
----------------------------------------------------------------------------------------------------

-- |   (, ,  )
data Indicator = NoIndicator | ShortIndicator | LongIndicator   deriving (Eq)

bytes_per_sec = 1*mb_  -- Typical (de)compression speed

-- |  ,     :)
select_indicator command total_bytes  = case (opt_indicator command)
  of "0"                                  ->  NoIndicator      --  "-i" -  !
     _ | total_bytes < bytes_per_sec*100  ->  ShortIndicator   --   ,      100  (        )
       | otherwise                        ->  LongIndicator    --     ,    100 

-- | `str`    - 
if_indicator ui_state str  = case (indicator ui_state)
  of NoIndicator     ->  ""
     _               ->  str

-- |       
percents ui_state       = case (indicator ui_state)
  of NoIndicator     ->  ""
     ShortIndicator  ->  right_justify 3 (ratio2 (bytes ui_state) (total_bytes ui_state)) ++ "%"
     LongIndicator   ->  right_justify 5 (ratio3 (bytes ui_state) (total_bytes ui_state)) ++ "%"

-- |     ,      
back_percents ui_state  =  case (indicator ui_state)
  of NoIndicator     ->  ""
     ShortIndicator  ->  "\b\b\b\b"
     LongIndicator   ->  "\b\b\b\b\b\b"

-- |     
clear_percents ui_state =  case (indicator ui_state)
  of NoIndicator     ->  ""
     ShortIndicator  ->  "\b\b\b\b    "
     LongIndicator   ->  "\b\b\b\b\b\b      "



----------------------------------------------------------------------------------------------------
----    (" ?"  ..) ------------------------------------------
----------------------------------------------------------------------------------------------------

-- |      
ask question ref_answer answer_on_u =  do
  old_answer <- readIORef ref_answer
  new_answer <- case old_answer of
                  "a" -> return old_answer
                  "u" -> return old_answer
                  "s" -> return old_answer
                  _   -> do putStr$ "\n  "++question++" (y/n/a/u/s/q)? "
                            hFlush stdout
                            answer  <-  getLine >>== strLower
                            when (answer=="q") $ do
                              registerError TERMINATED
                            writeIORef ref_answer answer
                            return answer
  case new_answer of
    "u" -> return answer_on_u
    _   -> return (new_answer `elem` ["y","a"])

-- |   
askOverwrite filename  =  ask ("Overwrite " ++ str2terminal filename)


----------------------------------------------------------------------------------------------------
---- /     -------------------------------------------------------------
----------------------------------------------------------------------------------------------------

-- |     
uiPrintArcComment arcComment = do
  when (arcComment>"") $ do
    putStrLn arcComment

-- |  stdin   
uiInputArcComment = do
  putStrLn "Enter archive comment, ending with \".\" on separate line:"
  hFlush stdout
  let go xs = do line  <-  getLine
                 if line/="."
                   then go (line:xs)
                   else return xs
  go []  >>==  reverse  >>==  join_with "\n"


----------------------------------------------------------------------------------------------------
----     /     --------------------
----------------------------------------------------------------------------------------------------

-- |      -    !!!
diffTimes (TOD sa pa) (TOD sb pb)  =  i(sa - sb) + (i(pa-pb) / 1e12)

-- | ,  ,   ..: "1.234.567"
show3 :: (Show a) => a -> [Char]
show3 = reverse.xxx.reverse.show
          where xxx (a:b:c:d:e) = a:b:c:'.': xxx (d:e)
                xxx a = a

-- |Format percent ratio with 2 digits
ratio2 count 0     =  "0"
ratio2 count total =  show$ count*100 `div` total

-- |Format percent ratio with 2+1 digits
ratio3 count 0     =  "0"
ratio3 count total =  case (show$ count*1000 `div` total) of
                        [digit]  -> "0." ++ [digit]
                        digits   -> init digits ++ ['.', last digits]

-- |     ,   
show_ratio cmd bytes cbytes =
  ""             ++ show3 (if (cmdType cmd == ADD_CMD) then bytes else cbytes) ++
   " => "        ++ show3 (if (cmdType cmd == ADD_CMD) then cbytes else bytes) ++ " bytes. " ++
   "Ratio "      ++ ratio3 cbytes bytes ++ "%"

-- | ,       (   )
statMark label = do
  command  <-  readIORef ref_command
  when (opt_debug command) $ do
    putStr$   label
    real_time <- return_real_time
    putStrLn$ ": " ++ real_time ++ " secs"

-- |,         
return_real_secs = do
  start_time    <- readIORef refStartArchiveTime
  current_time  <- getClockTime
  return$ diffTimes current_time start_time

-- | ,            
return_real_time = do
  real_secs <- return_real_secs
  return$ showFFloat (Just 2) real_secs ""


-- |    -     ,     
data ActionState = ActionState { old_time  :: !ClockTime
                               , old_param :: !String
                               }

-- |  `action`   `param`      
--    `maxdiff`   `param`  
do_if old_state new_time maxdiff action param = do
  let refresh  =  case (old_state) of
                    Nothing  ->  True
                    Just x   ->  (diffTimes new_time (old_time x)) >= maxdiff
                                  && (param /= old_param x)
  if refresh
    then do action param
            return$ Just$ ActionState new_time param
    else    return$ old_state

{-# NOINLINE diffTimes #-}
{-# NOINLINE show3 #-}
{-# NOINLINE ratio2 #-}
{-# NOINLINE ratio3 #-}
{-# NOINLINE show_ratio #-}
{-# NOINLINE statMark #-}
{-# NOINLINE do_if #-}


----------------------------------------------------------------------------------------------------
----- External functions ---------------------------------------------------------------------------
----------------------------------------------------------------------------------------------------

-- |Set console title
setConsoleTitle title = do
  withCString title c_SetConsoleTitle

-- |Set console title (external)
foreign import ccall unsafe "Environment.h EnvSetConsoleTitle"
   c_SetConsoleTitle :: Ptr CChar -> IO ()

-- |Reset console title
foreign import ccall unsafe "Environment.h EnvResetConsoleTitle"
   resetConsoleTitle :: IO ()


{-
   UI:
  -  ,    /   
        UI:
        ui_PROCESS pipe = do
          (StartCommand cmd) <- receiveP pipe
            (StartArchive cmd) <- receiveP pipe
              (StartFile fi fi) <- receiveP pipe
                (UnpackedData n) <- receiveP pipe
                (CompressedData n) <- receiveP pipe
            (EndArchive) <- receiveP pipe
          (EndCommand) <- receiveP pipe
         (EndProgram) <- receiveP pipe
           UI  SampleVar
-}
