---------------------------------------------------------------------------------------------------
----      /  .                          ----
---------------------------------------------------------------------------------------------------
module Cmdline where

import Prelude hiding (catch)
import Control.Exception
import Control.Monad
import Data.Array
import Data.Char
import Data.IORef
import Data.List
import Data.Maybe
import Foreign.C.Types
import System.Time
import System.Directory
import System.Environment

import Utils
import Files
import Errors
import FileInfo
import Compress


-- |  
data Command = Command {
    cmd_name                 :: !String             --  
  , cmd_arcspec              ::  String             --  
  , cmd_arclist              ::  [FilePath]         --         ( , ) 
  , cmd_arcname              ::  FilePath           --      (    wildcards  cmd_arcspec        )
  , cmd_arc_filter           :: !(FileInfo -> Bool) --      
  , cmd_filespecs            :: ![String]           --     
  , cmd_added_arcnames       :: !(IO [FilePath])    --   ,     ( "j")
  , cmd_diskfiles            :: !(IO [FileInfo])    --   ,      (  / )
                                                    -- :
  , opt_scan_subdirs         :: !Bool               --     ?
  , opt_add_dir              :: !Bool               --        ,   ?
  , opt_arc_basedir          :: !String             --      
  , opt_disk_basedir         :: !String             --      
  , opt_group_dir            :: ![Grouping]         --       
  , opt_group_data           :: ![Grouping]         --      -
  , opt_data_compressor      :: !UserCompressor     --      
  , opt_dir_compressor       :: !UserCompressor     --       
  , opt_arccmt_file          :: !String             --   ,    (  )   
  , opt_include_dirs         :: !(Maybe Bool)       --      ? (// )
  , opt_indicator            :: !String             --      ("0" - , "1" -   , "2" -        )
  , opt_overwrite            :: !String             --     ? (//   )
  , opt_sfx                  :: !String             --    SFX-,      ("-" - ,   , "--" -  )
  , opt_keep_time            :: !Bool               --    mtime    ?
  , opt_time_to_last         :: !Bool               --    mtime   mtime     ?
  , opt_keep_broken          :: !Bool               --     ,   ?
  , opt_test                 :: !Bool               --      ?
  , opt_lock_archive         :: !Bool               --        ?
  , opt_match_with           :: !(PackedFilePath -> FilePath)  --      fpBasename  fpFullname
  , opt_append               :: !Bool               --         ?
  , opt_noarcext             :: !Bool               --         ?
  , opt_nodir                :: !Bool               --         ( )?
  , opt_debug                :: !Bool               --     ?
  , opt_cache                :: !Int                --      
  , opt_update_type          :: !Char               --      (a/f/u/s)
  , opt_x_include_dirs       :: !Bool               --       (  /)?
  , opt_match_excluded       :: !(PackedFilePath -> Bool)  -- ,  ,     "-x"
  , opt_sort_order           :: !String             --       
  , opt_find_group           :: !(FileInfo -> Int)  --   ,   FileInfo    ( arc.groups)   
  , opt_groups_count         :: !Int                --     (`opt_find_group`     0..opt_groups_count-1)
  , opt_find_type            :: !(FileInfo -> Int)  --   ,   FileInfo     (   `opt_data_compressor`)   
  , opt_types_count          :: !Int                --      (`opt_find_type`     0..opt_types_count-1)

  , ref_ask_overwrite        :: !(IORef String)     --        ("a" -  , "s" -  )
  }

-- | ,  
optionsList =
   [OPTION "cfg-"  "config-"         ("ignore options in " ++ aCONFIG_FILE ++ " and %" ++ aCONFIG_ENV_VAR)
   ,OPTION "r"     "recursive"        "recursively collect files"
   ,OPTION "f"     "freshen"          "freshen files"
   ,OPTION "u"     "update"           "update files"
   ,OPTION ""      "sync"             "synchronize archive and disk contents"
   ,OPTION "o"     "overwrite"        "existing files overwrite MODE (+/-/p)"
   ,OPTION "y"     "yes"              "answer Yes to all queries"
   ,OPTION "x"     "exclude"          "exclude FILESPECS from operation"
   ,OPTION "ap"    "arcpath"          "base DIR in archive"
   ,OPTION "dp"    "diskpath"         "base DIR on disk"
   ,OPTION "m"     "method"           "compression METHOD"
   ,OPTION "dm"    "dirmethod"        "compression METHOD for archive directory"
   ,OPTION "md"    "dictionary"       "set compression dictionary to N bytes"
   ,OPTION "ms"    "StoreCompressed"  "store already compressed files"
   ,OPTION "ds"    "sort"             "sort files in ORDER"                      -- to do:    OptArg
   ,OPTION "g"     "groups"           "name of groups FILE"                      -- to do:    OptArg
   ,OPTION "s"     "solid"            "GROUPING for solid compression"           -- to do:    OptArg
   ,OPTION "t"     "test"             "test files after archiving"
   ,OPTION "kb"    "keepbroken"       "keep broken extracted files"
   ,OPTION "k"     "lock"             "lock archive"
   ,OPTION "sfx"   ""                ("add sfx MODULE (\""++aDEFAULT_SFX++"\" by default)")  -- to do:    OptArg
   ,OPTION "z"     "arccmt"           "read archive comment from FILE or stdin"  -- to do:    OptArg
   ,OPTION "i"     "indicator"        "select progress indicator TYPE (0/1/2)"   -- to do:    OptArg
   ,OPTION "ad"    "adddir"           "add arcname to extraction path"
   ,OPTION "ag"    "autogenerate"     "autogenerate archive name with FMT"       -- to do:    OptArg
   ,OPTION ""      "noarcext"         "don't add default extension to archive name"
   ,OPTION "tk"    "keeptime"         "keep original archive time"
   ,OPTION "tl"    "timetolast"       "set archive time to latest file"
   ,OPTION "fn"    "fullnames"        "match with full names"
   ,OPTION ""      "append"           "add new files to the end of archive"
   ,OPTION ""      "dirs"             "add empty dirs to archive"
   ,OPTION "ed"    "nodirs"           "don't add empty dirs to archive"
   ,OPTION ""      "cache"            "use N bytes for read-ahead cache"
   ,OPTION ""      "LimitCompMem"     "limit memory usage for compression to N bytes"
   ,OPTION ""      "LimitDecompMem"   "limit memory usage for decompression to N bytes"
   ,OPTION ""      "nodir"            "don't write archive directories"
   ,OPTION ""      "nodata"           "don't store data in archive"
   ,OPTION ""      "crconly"          "save/check CRC, but don't store data"
   ,OPTION ""      "debug"            "print debug info"
   ]

-- | ,           
aPREFFERED_OPTIONS = words "dictionary StoreCompressed sfx"

-- |   
aLL_COMMANDS = words "create a f m u j y s k c cw d e x t l v"

-- |     :  , ,   
data CmdType = ADD_CMD | EXTRACT_CMD | TEST_CMD | LIST_CMD   deriving (Eq)
cmdType "t" = TEST_CMD
cmdType "e" = EXTRACT_CMD
cmdType "x" = EXTRACT_CMD
cmdType "l" = LIST_CMD
cmdType "v" = LIST_CMD
cmdType  _  = ADD_CMD

-- |HELP,      
aHELP = "Usage: Arc command [options...] archive [files... @listfiles...]\n" ++
        "Commands: " ++ join_with ", " aLL_COMMANDS ++ "\n" ++ optionsHelp

-- |   -   
data Grouping = GroupNone                  --   
                                           --  
              | GroupByExt                 --    
              | GroupBySize      FileSize  --      
              | GroupByBlockSize Int       --       ( - ,   BWT  ST)
              | GroupByNumber    Int       --    
              | GroupAll                   --   

-- |     
aDEFAULT_DIR_GROUPING  =  GroupByNumber (20*1000)

-- |         -  , .. "*"
aDEFAULT_FILESPECS = [reANY_FILE]

-- |    
aDEFAULT_ARC_EXTENSION = ".arc"

-- |   SFX  
aDEFAULT_SFX_EXTENSION = ".exe"

-- |        "-og"
aDEFAULT_GROUPS_FILE = "arc.groups"

-- |SFX-,   
aDEFAULT_SFX = "arc.sfx"

-- |  ( ,   )
aCONFIG_FILE = "arc.ini"

-- | ,  ,   
aCONFIG_ENV_VAR = "FREEARC"

-- | ,     solid- (  )
aDEFAULT_SOLID_SORT_ORDER = "gen"


-- |             Command.
--     ,  ,     .
--   " ; ",  "a archive -r ; t archive ; x archive"
parseCmdline cmdline  =  (`mapMaybeM` split ";" (map cmdline2str cmdline)) $ \one_command -> do
  --     -  help  
  if one_command==[] then do
      putStr aHELP
      return Nothing
    else do

  --      ,    arc.ini     FREEARC
  let options    =  takeWhile (/="--") one_command
      no_configs =  "-cfg-" `elem` options  ||  "--config-" `elem` options
  add_options  <-
    if no_configs
      then return []  --      "-cfg-"
      else do
        cfgfile      <-  findConfigFile aCONFIG_FILE
        cfg_options  <-  case cfgfile of
                           "" -> return ""          -- - 
                           _  -> readFile cfgfile   --    -
        env_options  <-  getEnv aCONFIG_ENV_VAR  `catch`  (\e -> return "")
        return$ words cfg_options ++ words env_options
  let args  =  add_options ++ one_command

  --   ,      " "
  (o, freeArgs)  <-  parseOptions args [] []
  --   ,  " "   -     
  case (freeArgs) of
    []     ->  registerError$ CMDLINE_NO_COMMAND args
    [cmd]  ->  registerError$ CMDLINE_NO_ARCSPEC args
    cmd:pure_arcspec:pure_filespecs -> do
                                 -- :       
      let scan_subdirs          =  findNoArg    o "recursive"
          orig_exclude_list     =  findReqList  o "exclude"
          freshen               =  findNoArg    o "freshen"
          update                =  findNoArg    o "update"
          sync                  =  findNoArg    o "sync"
          add_dir               =  findNoArg    o "adddir"
          arc_basedir           =  findReqArg   o "arcpath"    ""     .$ stripPathSeparatorAtTheEnd
          disk_basedir          =  findReqArg   o "diskpath"   ""     .$ stripPathSeparatorAtTheEnd
          grouping              =  findReqArg   o "solid"      "64mb" .$ parseSolidOption
          orig_data_compressor  =  findReqArg   o "method"     "3"    .$ decode_method
          dir_compressor        =  findReqArg   o "dirmethod"  defaultDirCompressor .$ decode_method
          store_compressed      =  findNoArg    o "StoreCompressed"
          orig_sort_order       =  findMaybeArg o "sort"
          group_file            =  findReqArg   o "groups"       ""
          overwrite             =  findReqArg   o "overwrite"    "p"
          autogenerate_arcname  =  findOptArg   o "autogenerate" "-"  ||| "%Y%m%d%H%M%S"
          cache                 =  findOptArg   o "cache"        "1m" ||| "16m" .$ parseMem .$ (`atLeast` aBUFFER_SIZE)   -- 1mb  , 16 mb c  "--cache"  ,      
          indicator             =  findOptArg   o "indicator"    "1"  ||| "0"   --   -i1; -i  -i0
          arccmt_file           =  findOptArg   o "arccmt"       (if cmd=="c"  then ""  else "--")   --  "c"   "y -z"
          orig_sfx              =  findOptArg   o "sfx"          (if cmd=="s"  then ""  else "--")   --  "s"   "y -sfx"
          include_dirs          =  findNoArgs   o "dirs" "nodirs"
          limitCompressionMem   =  findReqArg   o "limitCompMem"    "--"
          limitDecompressionMem =  findReqArg   o "limitDecompMem"  "-"
          dictionary            =  findReqArg   o "dictionary"      "0"
          yes                   =  findNoArg    o "yes"
          keep_time             =  findNoArg    o "keeptime"
          time_to_last          =  findNoArg    o "timetolast"
          test_files            =  findNoArg    o "test"
          keep_broken           =  findNoArg    o "keepbroken"
          orig_lock_archive     =  findNoArg    o "lock"
          match_with            =  if(findNoArg o "fullnames")  then fpFullname  else fpBasename
          append                =  findNoArg    o "append"
          noarcext              =  findNoArg    o "noarcext"
          crconly               =  findNoArg    o "crconly"
          nodata                =  findNoArg    o "nodata"
          nodir                 =  findNoArg    o "nodir"
          debug                 =  findNoArg    o "debug"
          group_dir             =  fst3 grouping
          group_data            =  snd3 grouping
          defaultDirCompressor  =  thd3 grouping ||| "4x:512k"   --    ,   ,     "-dm"

      -- ,       
      testOption "overwrite" "o" overwrite (words "+ - p")
      testOption "indicator" "i" indicator (words "0 1 2")

      --   SFX-,      
      sfx <- case orig_sfx of
               "" -> findConfigFile aDEFAULT_SFX    --   SFX   (arc.sfx  ,   )
               _  -> return orig_sfx
      when (sfx=="") $
        registerError$ GENERAL_ERROR$ "SFX module "++aDEFAULT_SFX++" is not found"

      --       /,    -ag
      current_time <- getClockTime
      let arcspec2 = if autogenerate_arcname /= "-"
                       then updateFilenameBasenamePrefix pure_arcspec (++ showtime autogenerate_arcname current_time)
                       else pure_arcspec

      --       ,         --noarcext
      let arcspec  =  addArcExtension noarcext arcspec2

      --    "-x"   - ("-x@listfile")  
      exclude_list  <-  replace_list_files orig_exclude_list  >>==  map translatePath

      --    
      let match_excluded  =  match_filespecs match_with exclude_list

      --    - ("@listfile")  
      listed_filespecs  <-  replace_list_files pure_filespecs  >>==  map translatePath

      --          d/j,    
      filespecs <- case listed_filespecs of
        [] | cmd `elem` (words "cw d j")  ->  registerError$ CMDLINE_NO_FILENAMES args
           | otherwise                    ->  return aDEFAULT_FILESPECS
        _                                 ->  return listed_filespecs

--    ---------------------------------------------------
      --  "-md"        RAR :)
      let dict = case dictionary of
                   [c] | isAlpha c     -> 64*kb*2^(ord c - ord 'a')
                   s   | all isDigit s -> (readInt s)*kb
                   s                   -> parseMem s

      --    -ms,      lzp    
      let process_ms compressor = if store_compressed && not (isStoring compressor)
                                    then compressor ++ [("$compressed", decode_compressor "lzp:4:h13")]
                                    else compressor

      --   ,   
      let climit = case limitCompressionMem of
                     "--" -> clipToMaxInt$ getPhysicalMemory `div` 4 * 3  --      3/4    ( )
                     "-"  -> maxBound::Int  --    
                     s    -> parseMem s     --       

      --   ,    
      let dlimit = case limitDecompressionMem of
                     "-"  -> maxBound::Int  --    
                     s    -> parseMem s     --       

      --    "--nodata",    
      --    "--crconly",    CRC  
      let data_compressor = if      nodata   then [("", [aFAKE_COMPRESSION])]
                            else if crconly  then [("", [aCRC_ONLY_COMPRESSION])]
                            else orig_data_compressor
                                   .$ compressorSetDictionary dict
                                   .$ process_ms
                                   .$ compressorLimitCompressionMem climit
                                   .$ compressorLimitDecompressionMem dlimit
-------------------------------------------------------------------------------------

--      (find_group)    (find_type) ------------------
      -- ,      ( arc.groups)  
      actual_group_file <- case group_file of
                             "" -> findConfigFile aDEFAULT_GROUPS_FILE    --      (arc.groups  ,   )
                             "-"-> return ""          --     "-g-"
                             _  -> return group_file  --      "-gFILENAME"

      --      
      group_strings  <-  if actual_group_file > ""
                             then readFile actual_group_file        --   
                                    >>== filelist2str               --   Unicode
                                    >>== lines                      --    
                                    >>== filter (not.match ";*")    --   
                                    >>== map translatePath          --   '\'  '/'
                             else return [reANY_FILE]     --     ,       
      --  ,     
      let group_predicates  =  map (match_FP fpBasename) group_strings
      --   ,    ,       .
      --  - "$default",    ,       
      let lower_group_strings = (map strLower group_strings) ++ ["$default"]
          default_group = "$default" `elemIndex` lower_group_strings .$ fromJust
      --  "PackedFilePath ->    arc.groups"
      let find_group    = findGroup group_predicates default_group

      --    ($text, $exe   ),     arc.groups
      let group_type_names = go "$binary" lower_group_strings  --   - "$binary"
          go t []     = []           --    ,   
          go t (x:xs) = case x of    --         ("$text", "$rgb"   )
                          '$':_ | x/="$default" -> x : go x xs
                          _                     -> t : go t xs
      --       `data_compressor`,     arc.groups
      let group_types =  map typeNum group_type_names
          typeNum t   =  t `elemIndex` (map fst data_compressor) `defaultVal` 0
      --  ,       ,   `data_compressor`
      let type_predicates  =  const False : map match_type [1..maximum group_types]
          match_type t     =  any_function$ concat$ zipWith (\a b->if a==t then [b] else []) group_types group_predicates
      --  "PackedFilePath ->     `data_compressor`"
      let find_type        =  findGroup type_predicates 0
-------------------------------------------------------------------------------------

      --   
      let update_type = case cmd of
            "f"            -> 'f'   --  "f":         ,    
            "u"            -> 'u'   --  "u":             
            _ | freshen    -> 'f'   --   "-f":     . 
              | update     -> 'u'   --   "-u":     . 
              | sync       -> 's'   --   "--sync":          
              | otherwise  -> 'a'   -- :                      

      --    ,    "-k"   "k"
      let lock_archive  =  orig_lock_archive || cmd=="k"

      --    ?      /
      let x_include_dirs  =  case include_dirs of
                               Just x  -> x   --     --dirs/--nodirs
                               _       -> -- ,        "e"
                                          (filespecs==aDEFAULT_FILESPECS) && (cmd/="e")

      --      
      let sort_order  =  case (orig_sort_order, group_data) of
            (Just "-", _)  -> ""                  --      "-",   
            (Just  x,  _)  -> x                   --      ,   
            (_, [GroupNone]) -> ""                --    solid- -  
            _  -> if isStoring data_compressor  ||  isFakeCompression data_compressor
                  then ""                         --  -m0/--nodata/--crconly -   
                  else aDEFAULT_SOLID_SORT_ORDER  --  -      solid-

      --       
      ref_ask_overwrite  <-  newIORef (if yes  then "a"  else " ")

      --      
      return$ Just$ Command {
          cmd_name                 = cmd
        , cmd_arcspec              = arcspec
        , cmd_arclist              = error "Using uninitialized cmd_arclist"
        , cmd_arcname              = error "Using uninitialized cmd_arcname"
        , cmd_arc_filter           = const True
        , cmd_filespecs            = filespecs
        , cmd_added_arcnames       = return []
        , cmd_diskfiles            = return []

        , opt_scan_subdirs         = scan_subdirs
        , opt_add_dir              = add_dir
        , opt_arc_basedir          = arc_basedir
        , opt_disk_basedir         = disk_basedir
        , opt_group_dir            = group_dir
        , opt_group_data           = group_data
        , opt_data_compressor      = data_compressor
        , opt_dir_compressor       = dir_compressor
        , opt_include_dirs         = include_dirs
        , opt_indicator            = indicator
        , opt_overwrite            = overwrite
        , opt_keep_time            = keep_time
        , opt_time_to_last         = time_to_last
        , opt_test                 = test_files
        , opt_keep_broken          = keep_broken
        , opt_match_with           = match_with
        , opt_append               = append
        , opt_noarcext             = noarcext
        , opt_nodir                = nodir
        , opt_debug                = debug
        , opt_cache                = cache
        , opt_update_type          = update_type
        , opt_x_include_dirs       = x_include_dirs
        , opt_match_excluded       = match_excluded
        , opt_sort_order           = sort_order
        , opt_find_group           = find_group . fiFilteredName
        , opt_groups_count         = length group_strings
        , opt_find_type            = find_type . fiFilteredName
        , opt_types_count          = maximum group_types + 1
        , opt_arccmt_file          = arccmt_file
        , opt_lock_archive         = lock_archive
        , opt_sfx                  = sfx

        , ref_ask_overwrite        = ref_ask_overwrite
        }

-- |,       
testOption fullname shortname option valid_values = do
  unless (option `elem` valid_values) $ do
    registerError$ INVALID_OPTION_VALUE fullname shortname valid_values

--           --noarcext,
--       
addArcExtension noarcext filespec =
  case ('.' `elem` (basenameOf filespec), noarcext) of
    (False, False)  ->  filespec ++ aDEFAULT_ARC_EXTENSION
    _               ->  filespec

-- |   - ("@listfile")  
replace_list_files filespecs  =  concatMapM replace_list_file filespecs
  where replace_list_file filespec =
          case (start_from "@" filespec) of
            Just filename  ->  readFile filename >>== filelist2str >>== lines
            _              ->  return [filespec]

-- |   "-s"
parseSolidOption opt =
  case (split ';' opt) of
    []        ->  ([aDEFAULT_DIR_GROUPING], [GroupAll], "")   -- "-s"   -       
    ["-"]     ->  ([aDEFAULT_DIR_GROUPING], [GroupNone], "")  -- "-s-"  -,     
    ["7z"]    ->  ([GroupAll],  [GroupAll], "")               -- "-s=7z"        -     
    ["cab"]   ->  ([GroupAll],  [GroupAll],  "0")   --  -dm0  -- "-s=cab"       -     
    ["zip"]   ->  ([GroupAll],  [GroupNone], "0")   --  -dm0  -- "-s=zip"   -     ,    
    ["arj"]   ->  ([GroupNone], [GroupNone], "0")   --  -dm0  -- "-s=arj"   -       
    [dat]     ->  ([aDEFAULT_DIR_GROUPING], parse dat, "")    -- "-sXXX"     -,   
    [dir,dat] ->  (parse dir, parse dat, "")                  -- "-sXXX;YYY"     ,   -
  where
    --    :
    -- "-s/-se/-s10m/-s100f" -  / / 10 / 100 , 
    parse = map parse1 . recursive (spanBreak (\c -> isDigit c || c=='e'))
    parse1 s = case s of
                ""  -> GroupAll
                "e" -> GroupByExt
                _   -> case (parseNumber s 'f') of
                         (num, 'b') -> GroupBySize num
                         (1,   'f') -> GroupNone
                         (num, 'f') -> GroupByNumber (i num)


{-# NOINLINE cmdType #-}
{-# NOINLINE testOption #-}
{-# NOINLINE replace_list_files #-}


----------------------------------------------------------------------------------------------------
----         -----------------------------------------------
----------------------------------------------------------------------------------------------------

-- |  -  ,  ,  
data Option = OPTION String String String

-- |  -    "-",  -  "--"
data OptType  =  SHORT | LONG

-- |   : //
data ParamType  =  ParamNo | ParamReq | ParamOpt

-- |"" ,         
optionsDict  =  foldr compileOption [] optionsList
  where compileOption (OPTION short long description)  =  compile short . compile ('-':long)
          where --         `name`,   
                compile name  =  case (name, paramName description) of
                                     ("",  _      )  ->  id                                 --   -    :)
                                     ("-", _      )  ->  id                                 --   -    :)
                                     (_,   Nothing)  ->  ((name, long|||short, ParamNo):)   --   
                                     (_,   Just _ )  ->  ((name, long|||short, ParamReq):)  --   

-- |   
optionsHelp  =  unlines table
  where (ss,ls,ds)     = (unzip3 . map fmtOpt) optionsList
        table          = zipWith3 paste (sameLen ss) (sameLen ls) ds
        paste x y z    = "  " ++ x ++ "  " ++ y ++ "  " ++ z
        sameLen xs     = flushLeft ((maximum . map length) xs) xs
        flushLeft n    = map (left_justify n)
          --   " ", " ",   
        fmtOpt (OPTION short long description)  =  (format short "" description, format ('-':long) "=" description, description)
          --    `name`        
        format name delim description  =  case (name, paramName description) of
                                            ("",  _         )  ->  ""
                                            ("-", _         )  ->  ""
                                            (_,   Nothing   )  ->  "-"++name
                                            (_,   Just aWORD)  ->  "-"++name++delim++aWORD

-- |   ,      
paramName descr =
  case filter (all isUpper) (words descr)
    of []      -> Nothing      --    UPPERCASED 
       [aWORD] -> Just aWORD   --   UPPERCASED ,   
       _       -> error$ "option description \""++descr++"\" contains more than one uppercased word"

-- |  ,      " "
parseOptions []          options freeArgs  =  return (reverse options, reverse freeArgs)
parseOptions ("--":args) options freeArgs  =  return (reverse options, reverse freeArgs ++ args)

parseOptions (('-':option):args) options freeArgs = do
  let check (prefix, _, ParamNo)  =  (option==prefix)
      check (prefix, _, _)        =  (start_from prefix option /= Nothing)
  let accept (prefix, name, haveParam)  =  return (name, tryToSkip "=" (tryToSkip prefix option))
      unknown                           =  registerError$ CMDLINE_UNKNOWN_OPTION ('-':option)
      ambiguous variants                =  registerError$ CMDLINE_AMBIGUOUS_OPTION ('-':option) (map (('-':).fst3) variants)
  newopt <- case (filter check optionsDict) of
              [opt] -> accept opt  --  
              []    -> unknown     --  
              xs    -> --          
                       case (filter ((`elem` aPREFFERED_OPTIONS) . snd3) xs) of
                         [opt] -> accept opt        --  
                         []    -> ambiguous xs      --  ,     
                         xs    -> ambiguous xs      --      !
  parseOptions args (newopt:options) freeArgs

parseOptions (arg:args) options freeArgs   =  parseOptions args options (arg:freeArgs)


-- |      `flag`.  : findReqList opts "exclude"
findReqList ((name, param):flags) flag  | name==flag  =  param: findReqList flags flag
findReqList (_:flags) flag                            =  findReqList flags flag
findReqList [] flag                                   =  []

-- |     `flag`,    -    `deflt`
findReqArg options flag deflt  =  last (deflt : findReqList options flag)

-- |     
findOptArg = findReqArg

-- |     `flag`,    - Nothing
findMaybeArg options flag  =  case findReqList options flag
                                of [] -> Nothing
                                   xs -> Just (last xs)

-- | True,         `flag`
findNoArg options flag  =  case findReqList options flag
                                of [] -> False
                                   _  -> True

-- | Just True,         `flag1`,
--          Just False,         `flag2`,
--          Nothing,    ,  
findNoArgs options flag1 flag2  =  case filter (\(o,_) -> o==flag1||o==flag2) options
                                     of [] -> Nothing
                                        xs -> Just (fst (last xs) == flag1)

{-# NOINLINE optionsDict #-}
{-# NOINLINE optionsHelp #-}
{-# NOINLINE parseOptions #-}
{-# NOINLINE findReqList #-}
{-# NOINLINE findReqArg #-}
{-# NOINLINE findMaybeArg #-}
{-# NOINLINE findNoArg #-}
{-# NOINLINE findNoArgs #-}

----------------------------------------------------------------------------------------------------
----- External functions ---------------------------------------------------------------------------
----------------------------------------------------------------------------------------------------

-- |   
foreign import ccall unsafe "Environment.h GetPhysicalMemory"
  getPhysicalMemory :: CUInt
