----------------------------------------------------------------------------------------------------
---- ,    CRC.                                                     ----
----   Compressor -   .                                            ----
----      ,    .               ----
----------------------------------------------------------------------------------------------------
module Compress where

import Control.Concurrent
import Control.Monad
import Data.Bits
import Data.Char
import Data.List
import Data.Maybe
import Data.Word
import Foreign.C.String
import Foreign.C.Types
import Foreign.Marshal.Alloc
import Foreign.Marshal.Pool
import Foreign.Ptr

import CompressionLib hiding (b,kb,mb,gb)
import Errors
import Utils
import Files
import qualified ByteStream


-- |      
type CompressionMethod  =  String

--  "",  ,    CompressionLib
aSTORING              = "storing"
aFAKE_COMPRESSION     = "fake"
aCRC_ONLY_COMPRESSION = "crc"

-- | ()  
isFakeMethod method  =  method==aFAKE_COMPRESSION  ||  method==aCRC_ONLY_COMPRESSION


-- |  ,    
type Compressor = [CompressionMethod]

-- | "storing" (-m0)
aNO_COMPRESSION = [] :: Compressor

-- | -  ,          - 
isFakeCompressor [method]  =  isFakeMethod method
isFakeCompressor _         =  False



-- |       
type UserCompressor = [(String,Compressor)]  --    "$text->m3t, $exe->m3x, $compressed->m0"

-- | -  Storing,       aNO_COMPRESSION    
isStoring ((_,compressor):xs)  =  compressor==aNO_COMPRESSION  &&  xs==[]

-- | - fake compression,           
isFakeCompression ((_,compressor):xs)  =  isFakeCompressor compressor  &&  xs==[]

-- | ,      `s`
findCompressor (s::String) (uc::UserCompressor) =
  case (lookup s uc) of
    Just x -> x                  --        -  
    _      -> snd (head uc)      --  -       (  )

-- |         
instance ByteStream.BufferData Compressor where
  write buf x  =  ByteStream.write buf (encode_compressor x)
  read  buf    =  ByteStream.read  buf  >>==  decode_compressor


----------------------------------------------------------------------------------------------------
-- Encode/decode compression method for parsing options/printing info about compression method -----
----------------------------------------------------------------------------------------------------

-- |Parse command-line option representing compression method.
--   (UserCompressor)    "  -> "
--             ,      
decode_method str  =  map (mapSnd decode_compressor) $
  case (strLower str) of
    "1" -> [("", "1x")]
    "2" -> [("", "2t"), ("$text", "2t")]
    "3" -> [("", "3x"), ("$text", "3t")]
    "4" -> [("", "4x"), ("$text", "4t")]
    "4r"-> [("", "4x"), ("$text", "4t:r")]
    "5" -> [("", "5x"), ("$text", "5t")]
    "6" -> [("", "6x"), ("$text", "6t")]
    "1p" -> [("", "1xp")]
    "2p" -> [("", "2tp"), ("$text", "2tp")]
    "3p" -> [("", "3xp"), ("$text", "3tp")]
    "4p" -> [("", "4xp"), ("$text", "4tp")]
    "4pr"-> [("", "4xp"), ("$text", "4tp:r")]
    "5p" -> [("", "5xp"), ("$text", "5tp")]
    "6p" -> [("", "6xp"), ("$text", "6tp")]
    _ -> [("", str)]

-- |  (  )
decode_compressor str  =
  --  -      ,  ,
  --         ( "1x" -> "lzp:8m:4:h13"),
  --        
  let decode  =  map decode_one_method . subst_methods
  in case (strLower str) of
       "0"             ->  aNO_COMPRESSION    --   "-m0"
       "storing"       ->  aNO_COMPRESSION    --        :)
       '0':'=':method  ->  decode method      --   "-m0=..."
       method          ->  decode method      --   "-m..."

-- |    ,  "1x" -> "lzp:8m:4:h13"
subst_methods method =
  case (split ',' method) of
    [_]      ->  if result==method   then [result]   else subst_methods result   --          :)
    methods  ->  concatMap subst_methods methods   --    ,  ,      

  where --  ""  (,   ),  ,      
        result         =  new_head++params
        (head,params)  =  break (==':') method
        new_head = case (head) of
          "0"  -> "storing"
          "1x" -> "lzp:8m:4:h13"          ;    "1xp" -> "1x"
          "2x" -> "lzma:2m:fast:4"        ;    "2xp" -> "exe,lzp:8m:64:h13,m2x"
          "3x" -> "lzma:8m:fast:32"       ;    "3xp" -> "exe,lzp:8m:64:h15,m3x"
          "4x" -> "lzma:8m:normal:32"     ;    "4xp" -> "exe,lzp:8m:64:h18,m4x"
          "5x" -> "lzma:16m:max:64"       ;    "5xp" -> "exe,lzp:16m:64:h18,m5x"
          "6x" -> "lzma:32m:max:bt4b:128" ;    "6xp" -> "exe,lzp:32m:64:h20,m6x"

          "1t" -> "1x"                    ;    "1tp" -> "1t"
          "2t" -> "grzip:m4:8m:32:h15"    ;    "2tp" -> "2t"
          "3t" -> "ppmd:10:48m"           ;    "3tp" -> "lzp:8m:080:h18,m3t"
          "4t" -> "ppmd:14:96m"           ;    "4tp" -> "lzp:8m:120:h18,m4t"
          "5t" -> "ppmd:18:160m:r"        ;    "5tp" -> "lzp:8m:160:h18,m5t"
          "6t" -> "ppmd:24:384m:r"        ;    "6tp" -> "lzp:8m:200:h18,m6t"

          ['x',c] | isDigit c      ->  [c,'x'] --    "-mx7"    7-zip
          'm':x   | any isDigit x  ->  x       --      "-m0=lzp,m3"
          x                        ->  x       --        :)

-- |    
decode_one_method  =  canonizeCompressionMethod


-- |Human-readable description of compression method
encode_method uc  =  ", using " ++ join_with ", " (map encode_one_method uc)
encode_one_method (text,compressor)  =  between text " => " (encode_compressor compressor)
encode_compressor compressor  =  join_with "," compressor ||| "storing"


-- |,      /
showMem mem = if mem `mod` gb==0 || mem `div` mb>=10000   then show( mem             `div` gb)++"gb"
         else if mem `mod` mb==0 || mem `div` kb>=10000   then show((mem+mb `div` 2) `div` mb)++"mb"
         else if mem `mod` kb==0 || mem `div`  b>=10000   then show((mem+kb `div` 2) `div` kb)++"kb"
         else                                                  show( mem )                    ++"b"

{-
showMem = showMem' [(gb,'gb'),(mb,'mb'),(kb,'kb'),(b,'b')]

showMem' ((val,str):xs@(nextval,_)) mem =
  if mem `mod` val==0 || mem `div` nextval>=10
    then show((mem+val`div` 2) `div` val)++str
    else ss mem xs
-}

{-# NOINLINE decode_method #-}
{-# NOINLINE decode_compressor #-}
{-# NOINLINE decode_one_method #-}
{-# NOINLINE subst_methods #-}
{-# NOINLINE showMem #-}



----------------------------------------------------------------------------------------------------
----- (De)compression of data stream                                                           -----
----------------------------------------------------------------------------------------------------

{-
compress   comprMethod reader writer      -  
decompress comprMethod reader writer      -  

  comprMethod :: CompressionMethod -  
  reader buf size -      `buf`  `size`
                     0   -  
                               <0  -  () (     )
                               >0  - -  
  writer buf size -   
                     <0  -  () (     )
                               >=0 -  
                            "",  
                      ()         
       ()
-}

-- |     
freearcCompress      method | aSTORING ==  method =  copy_data
freearcCompress      method | isFakeMethod method =  eat_data
freearcCompress      method                       =  checkingCtrlBreak (CompressionLib.compress method)

-- |     
freearcDecompress    method | aSTORING ==  method =  copy_data
freearcDecompress    method | isFakeMethod method =  undecompressible   --       
freearcDecompress    method                       =  checkingCtrlBreak (CompressionLib.decompress method)

-- |       (BWT/ST/LZP)   0
freearcGetBlockSize  method | aSTORING ==  method =  0
freearcGetBlockSize  method | isFakeMethod method =  0
freearcGetBlockSize  method                       =  CompressionLib.getBlockSize method

-- |   CompressionLib,  method,  ,  method::COMPRESSION_METHOD
liftSetter action  method | aSTORING ==  method   =  method
liftSetter action  method | isFakeMethod method   =  method
liftSetter action  method                         =  action method

-- |/  ,      /
freearcSetDictionary          =  liftSetter . CompressionLib.setDictionary
freearcLimitDictionary        =  liftSetter . CompressionLib.limitDictionary
freearcLimitCompressionMem    =  liftSetter . CompressionLib.limitCompressionMem
freearcLimitDecompressionMem  =  liftSetter . CompressionLib.limitDecompressionMem

-- |  /    / 
--    ,   COMPRESSOR
compressorSetDictionary          = map . mapSnd . map . freearcSetDictionary
compressorLimitCompressionMem    = map . mapSnd . map . freearcLimitCompressionMem
compressorLimitDecompressionMem  = map . mapSnd . map . freearcLimitDecompressionMem

-- | Haskell' ,   ,   
-- ,    /  
checkingCtrlBreak action read write = do
  let checked f buf size = do aProgramTerminated <- val programTerminated
                              if aProgramTerminated
                                then return (-1)
                                else f buf size
  action (checked read) (checked write)

-- |    (-m0)
copy_data reader writer = do
  allocaBytes aBUFFER_SIZE $ \buf -> do  --  `alloca`,       
    let go = do
          len <- reader buf aBUFFER_SIZE
          if (len>0)
            then do result <- writer buf len
                    if (result>=0)
                      then go
                      else return result  --   ,   /   
            else return len  --  0,   ,   ,   /   
    go  --  

-- | ,   ,  CRC     ;)
eat_data reader writer = do
  allocaBytes aBUFFER_SIZE $ \buf -> do  --  `alloca`,       
    let go = do
          len <- reader buf aBUFFER_SIZE
          if (len>0)
            then go
            else return len   --  0,   ,   ,   /   
    go  --  

undecompressible reader writer = do
  return (-1)   --   ,    (FAKE/CRC_ONLY)   

{-# NOINLINE checkingCtrlBreak #-}
{-# NOINLINE copy_data #-}
{-# NOINLINE eat_data #-}


----------------------------------------------------------------------------------------------------
----- CRC calculation ------------------------------------------------------------------------------
----------------------------------------------------------------------------------------------------

-- |CRC 
type CRC  = CUInt
aINIT_CRC = 0xffffffff  :: CRC
updateCRC addr len  =  c_updateCRC addr (i len)
finishCRC = (aINIT_CRC `xor`)
-- | CRC   
calcCRC addr len  =  updateCRC addr len aINIT_CRC  >>==  finishCRC


----------------------------------------------------------------------------------------------------
----- External functions ---------------------------------------------------------------------------
----------------------------------------------------------------------------------------------------

-- |CRC-32 calculation
foreign import ccall unsafe "Environment.h UpdateCRC"
   c_updateCRC :: Ptr CChar -> CUInt -> CRC -> IO CRC
