#include <stdlib.h>
#include <stdio.h>
#include <io.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <string.h>

#include "../Common.h"
#include "Compression.h"

//     
int decompress (char *method, INOUT_FUNC *read_f, INOUT_FUNC *write_f)
{
  COMPRESSION_METHOD *compressor = ParseCompressionMethod (method);
  if (compressor){
    int result = compressor->decompress (read_f, write_f);
    delete compressor;
    return result;}
  else
    return FREEARC_ERRCODE_INVALID_COMPRESSOR;
}

//            
int DecompressWithHeader (INOUT_FUNC *read_f, INOUT_FUNC *write_f)
{
  char method [MAX_METHOD_STRLEN];
  for (int i=0; i<MAX_METHOD_STRLEN; i++)
  {
    //    ,      
    read_f ( &method[i], 1);
    if (method[i]=='\0')
      return decompress (method, read_f, write_f);
  }
  return FREEARC_ERRCODE_INVALID_COMPRESSOR;  //   ,    MAX_METHOD_STRLEN       '\0'
}

// Callback-   ()  
void *readPtr;    //    
int   readLeft;   //       
int ReadMem (void *buf, int size)
{
  int read_bytes = readLeft<size ? readLeft : size;
  memcpy (buf, readPtr, read_bytes);
  readPtr   = (uint8*)readPtr+read_bytes;
  readLeft -= read_bytes;
  return read_bytes;
}

// Callback-   ()  
void *writePtr;    //    
int   writeLeft;   //       
int WriteMem (void *buf, int size)
{
  if (size>writeLeft)  return FREEARC_ERRCODE_OUTBLOCK_TOO_SMALL;
  memcpy (writePtr, buf, size);
  writePtr   = (uint8*)writePtr+size;
  writeLeft -= size;
  return size;
}

//    ,       outputSize .
//      ,    
int DecompressMem (char *method, void *input, int inputSize, void *output, int outputSize)
{
  readPtr=input, readLeft=inputSize, writePtr=output, writeLeft=outputSize;
  int result = decompress (method, ReadMem, WriteMem);
  return result<0 ? result : outputSize-writeLeft;
}

//    ,       outputSize .
//      ,    
int DecompressMemWithHeader (void *input, int inputSize, void *output, int outputSize)
{
  readPtr=input, readLeft=inputSize, writePtr=output, writeLeft=outputSize;
  int result = DecompressWithHeader (ReadMem, WriteMem);
  return result<0 ? result : outputSize-writeLeft;
}

#ifndef FREEARC_DECOMPRESS_ONLY

//     
int compress (char *method, INOUT_FUNC *read_f, INOUT_FUNC *write_f)
{
  COMPRESSION_METHOD *compressor = ParseCompressionMethod (method);
  if (compressor){
    int result = compressor->compress (read_f, write_f);
    delete compressor;
    return result;}
  else
    return FREEARC_ERRCODE_INVALID_COMPRESSOR;
}

//            
int CompressWithHeader (char *method, INOUT_FUNC *read_f, INOUT_FUNC *write_f)
{
  COMPRESSION_METHOD *compressor = ParseCompressionMethod (method);
  if (compressor){
    char canonical_method [MAX_METHOD_STRLEN];
    compressor->ShowCompressionMethod (canonical_method);
    int result = write_f (canonical_method, strlen(canonical_method)+1);
    if (result>=0) result = compressor->compress (read_f, write_f);
    delete compressor;
    return result;}
  else
    return FREEARC_ERRCODE_INVALID_COMPRESSOR;
}

//    ,       outputSize .
//      ,    
int CompressMem (char *method, void *input, int inputSize, void *output, int outputSize)
{
  readPtr=input, readLeft=inputSize, writePtr=output, writeLeft=outputSize;
  int result = compress (method, ReadMem, WriteMem);
  return result<0 ? result : outputSize-writeLeft;
}

//    ,       outputSize .
//      ,    
int CompressMemWithHeader (char *method, void *input, int inputSize, void *output, int outputSize)
{
  readPtr=input, readLeft=inputSize, writePtr=output, writeLeft=outputSize;
  int result = CompressWithHeader (method, ReadMem, WriteMem);
  return result<0 ? result : outputSize-writeLeft;
}

//   canonical_method     in_method
int CanonizeCompressionMethod (char *method, char *canonical_method)
{
  COMPRESSION_METHOD *compressor = ParseCompressionMethod (method);
  if (compressor){
    compressor->ShowCompressionMethod (canonical_method);
    delete compressor;
    return FREEARC_OK;}
  else
    return FREEARC_ERRCODE_INVALID_COMPRESSOR;
}


#define Generate_Getter(GETTER)                                              \
  int GETTER (char *method)                                                  \
  {                                                                          \
    COMPRESSION_METHOD *compressor = ParseCompressionMethod (method);        \
    if (compressor){                                                         \
      int bytes = compressor->GETTER();                                      \
      delete compressor;                                                     \
      return bytes;}                                                         \
    else                                                                     \
      return FREEARC_ERRCODE_INVALID_COMPRESSOR;                             \
  }                                                                          \

#define Generate_Setter(SETTER)                                              \
  int SETTER (char *in_method, int bytes, char *out_method)                  \
  {                                                                          \
    COMPRESSION_METHOD *compressor = ParseCompressionMethod (in_method);     \
    if (compressor){                                                         \
      compressor->SETTER (bytes);                                            \
      compressor->ShowCompressionMethod (out_method);                        \
      delete compressor;                                                     \
      return FREEARC_OK;}                                                    \
    else                                                                     \
      return FREEARC_ERRCODE_INVALID_COMPRESSOR;                             \
  }                                                                          \

//   ,   /,     .
Generate_Getter(GetCompressionMem)
Generate_Getter(GetDecompressionMem)
Generate_Getter(GetDictionary)
Generate_Getter(GetBlockSize)

//   out_method   ,   
//     /  / 
Generate_Setter(SetCompressionMem)
Generate_Setter(SetDecompressionMem)
Generate_Setter(SetDictionary)
Generate_Setter(SetBlockSize)

//   out_method   , ,  ,
//    /   /  
Generate_Setter(LimitCompressionMem)
Generate_Setter(LimitDecompressionMem)
Generate_Setter(LimitDictionary)
Generate_Setter(LimitBlockSize)

#endif  // !defined (FREEARC_DECOMPRESS_ONLY)



// ****************************************************************************************************************************
//                *
// ****************************************************************************************************************************

CM_PARSER cmTable[MAX_COMPRESSION_METHODS];   // ,        
int cmCount = 0;                              // -   

//        
int AddCompressionMethod (CM_PARSER parser)
{
  cmTable[cmCount++] = parser;
  CHECK (cmCount <= elements(cmTable), ("\nINTERNAL ERROR: Overflow of compression methods table"));
}

//    COMPRESSION_METHOD,  ,     `method`
COMPRESSION_METHOD *ParseCompressionMethod (char* method)
{
  //        `parameters`,     
  char* parameters [MAX_PARAMETERS];
  char  local_method [MAX_METHOD_STRLEN];
  strncopy (local_method, method, sizeof (local_method));
  split (local_method, COMPRESSION_METHOD_PARAMETERS_DELIMITER, parameters);

  //         ,    `parameters`
  iterate_var (i, cmCount)  {
     COMPRESSION_METHOD *m = (*cmTable[i]) (parameters);
     if (m)  return m;
  }
  return NULL;   //         
}


// ***********************************************************************************************************************
//  ,         
// ***********************************************************************************************************************

//   param    -  ,   error=1
int parseInt  (char *param, int *error)
{
  int n=0;
  char c=*param;
  while (c>='0' && c<='9')  n=n*10+c-'0', c=*++param;
  if (c!='\0') *error=1;
  return n;
}

//  readInt,   param    b/k/m/g/^,      (  - '^', ..  )
int parseMem (char *param, int *error)
{
  int n=0;
  char c=*param;
  while (c>='0' && c<='9')  n=n*10+c-'0', c=*++param;
  switch (c)
  {
    case 'b':  return n;
    case 'k':  return n*kb;
    case 'm':  return n*mb;
    case 'g':  return n*gb;
    case '^':
    case '\0':  return 1<<n;
  }
  *error=1; return 0;
}

#ifndef FREEARC_DECOMPRESS_ONLY
//     
void showMem (int mem, char *result)
{
       if (mem%gb==0) sprintf (result, "%dgb", mem/gb);
  else if (mem%mb==0) sprintf (result, "%dmb", mem/mb);
  else if (mem%kb==0) sprintf (result, "%dkb", mem/kb);
  else                sprintf (result, "%db",  mem);
}
#endif

//   from  to,    len 
void strncopy( char *to, char *from, int len ) {
  for (int i = len; --i && *from; )     *to++ = *from++;
  *to = '\0';
}

//   str  ,   splitter.
//  -   str splitter   '\0'
//     result      str  + NULL ( argv)
void split (char *str, char splitter, char **result)
{
  *result++ = str;
  while (*str)
  {
    while (*str && *str!=splitter) str++;
    if (*str) {
      *str++ = '\0';
      *result++ = str;
    }
  }
  *result = NULL;
}



/*-------------------------------------------------*/
/*   STORING                       */
/*-------------------------------------------------*/
//  "()",     
int copy_data( INOUT_FUNC *read_f, INOUT_FUNC *write_f)
{
  char buf[BUFFER_SIZE]; int len;
  while ((len = read_f (buf, BUFFER_SIZE)) > 0) {
    write_f (buf, len);
  }
  return len;
}

//  
int STORING_METHOD::decompress (INOUT_FUNC *read_f, INOUT_FUNC *write_f)
{
  return copy_data (read_f, write_f);
}

#ifndef FREEARC_DECOMPRESS_ONLY

//  
int STORING_METHOD::compress (INOUT_FUNC *read_f, INOUT_FUNC *write_f)
{
  return copy_data (read_f, write_f);
}

//   buf[MAX_METHOD_STRLEN] ,    (,   parse_STORING)
void STORING_METHOD::ShowCompressionMethod (char *buf)
{
  sprintf (buf, "storing");
}

#endif  // !defined (FREEARC_DECOMPRESS_ONLY)

//    STORING_METHOD   NULL,     
COMPRESSION_METHOD* parse_STORING (char** parameters)
{
  if (strcmp (parameters[0], "storing") == 0
      &&  parameters[1]==NULL )
    //    - "storing"     ,    
    return new STORING_METHOD;
  else
    return NULL;   //    storing
}

static int STORING_x = AddCompressionMethod (parse_STORING);   //    STORING_METHOD
