----------------------------------------------------------------------------------------------------
----    .                                                              ----
----      ,    .               ----
----------------------------------------------------------------------------------------------------
module CompressionLib where

import Control.Concurrent
import Control.Exception
import Control.Monad
import Data.Bits
import Data.Char
import Data.List
import Data.Maybe
import Data.Word
import Foreign.C.String
import Foreign.C.Types
import Foreign.Marshal.Alloc
import Foreign.Marshal.Pool
import Foreign.Ptr
import System.IO.Unsafe

----------------------------------------------------------------------------------------------------
----- Main exported definitions --------------------------------------------------------------------
----------------------------------------------------------------------------------------------------

-- |Useful definitions for kilobytes, megabytes and so on
b:kb:mb:gb:_  = iterate (1024*) 1 :: [Int]

-- |Compress using callbacks
compress             :: Method -> INOUT_FUNC -> INOUT_FUNC -> IO Int
compress             = runWithMethod c_compress
-- |Decompress using callbacks
decompress           :: Method -> INOUT_FUNC -> INOUT_FUNC -> IO Int
decompress           = runWithMethod c_decompress
-- |Compress using callbacks and save method name in compressed output
compressWithHeader   :: Method -> INOUT_FUNC -> INOUT_FUNC -> IO Int
compressWithHeader   = runWithMethod c_CompressWithHeader
-- |Decompress data compressed with compressWithHeader (method name is read from compressed stream)
decompressWithHeader ::           INOUT_FUNC -> INOUT_FUNC -> IO Int
decompressWithHeader = run           c_DecompressWithHeader

-- |Compress memory block
compressMem             :: Method -> Ptr CChar -> Int -> Ptr CChar -> Int -> IO Int
compressMem             = withMethod c_CompressMem
-- |Decompress memory block
decompressMem           :: Method -> Ptr CChar -> Int -> Ptr CChar -> Int -> IO Int
decompressMem           = withMethod c_DecompressMem
-- |Compress memory block and save method name in compressed output
compressMemWithHeader   :: Method -> Ptr CChar -> Int -> Ptr CChar -> Int -> IO Int
compressMemWithHeader   = withMethod c_CompressMemWithHeader
-- |Decompress memory block compressed with compressMemWithHeader (method name is read from compressed data)
decompressMemWithHeader ::           Ptr CChar -> Int -> Ptr CChar -> Int -> IO Int
decompressMemWithHeader = c_DecompressMemWithHeader

-- |Return canonical representation of compression method
canonizeCompressionMethod :: Method -> Method
canonizeCompressionMethod = doWithMethod c_CanonizeCompressionMethod

-- |Returns memory used to compress/decompress, dictionary or block size of method given
getCompressionMem, getDecompressionMem, getDictionary, getBlockSize :: Method -> Int
getCompressionMem      =  getWithMethod c_GetCompressionMem
getDecompressionMem    =  getWithMethod c_GetDecompressionMem
getDictionary          =  getWithMethod c_GetDictionary
getBlockSize           =  getWithMethod c_GetBlockSize

-- |Set memory used to compress/decompress, dictionary or block size of method given
setCompressionMem, setDecompressionMem, setDictionary, setBlockSize :: Int -> Method -> Method
setCompressionMem      =  setWithMethod c_SetCompressionMem
setDecompressionMem    =  setWithMethod c_SetDecompressionMem
setDictionary          =  setWithMethod c_SetDictionary
setBlockSize           =  setWithMethod c_SetBlockSize

-- |Put upper limit to memory used to compress/decompress, dictionary or block size of method given
limitCompressionMem, limitDecompressionMem, limitDictionary, limitBlockSize :: Int -> Method -> Method
limitCompressionMem    =  setWithMethod c_LimitCompressionMem
limitDecompressionMem  =  setWithMethod c_LimitDecompressionMem
limitDictionary        =  setWithMethod c_LimitDictionary
limitBlockSize         =  setWithMethod c_LimitBlockSize


----------------------------------------------------------------------------------------------------
----- Internal auxiliary functions -----------------------------------------------------------------
----------------------------------------------------------------------------------------------------

-- |Run specified action with Haskell arguments converted to C equivalents
runWithMethod action method read_f write_f = do
  withCString method $ \c_method -> do
    run (action c_method) read_f write_f

-- |Execute C (de)compression routine `action` using read/write callbacks `read_f` & `write_f`
run action read_f write_f = do
  bracket (mkInOutFunc read_f) (freeHaskellFunPtr)$ \call_r -> do      -- convert Haskell routines
    bracket (mkInOutFunc write_f) (freeHaskellFunPtr)$ \call_w -> do   --   to C-callable routines
      action call_r call_w    -- COMPRESSION or DECOMPRESSION

withMethod action method inp insize outp outsize = do
  withCString method $ \c_method -> do
    action c_method inp insize outp outsize

getWithMethod c_get method  =  unsafePerformIO$ withCString method c_get

setWithMethod c_action bytes = doWithMethod (\x y -> c_action x bytes y)

doWithMethod c_action method = do
  unsafePerformIO $ do
    withCString method $ \c_method -> do
      allocaBytes 256  $ \c_out_method -> do
        ret <- c_action c_method c_out_method
        case ret of
          0 -> peekCString c_out_method
          _ -> fail$ "Unsupported compression method or error in parameters: "++method


{-# NOINLINE run #-}
{-# NOINLINE doWithMethod #-}

----------------------------------------------------------------------------------------------------
----- High-level compression/decompression routines working with callbacks -------------------------
----------------------------------------------------------------------------------------------------

-- |Compress using callbacks
foreign import ccall threadsafe  "Compression.h compress"
   c_compress             :: CMethod -> FunPtr INOUT_FUNC -> FunPtr INOUT_FUNC -> IO Int

-- |Decompress using callbacks
foreign import ccall threadsafe  "Compression.h decompress"
   c_decompress           :: CMethod -> FunPtr INOUT_FUNC -> FunPtr INOUT_FUNC -> IO Int

-- |Compress using callbacks and save method name in compressed output
foreign import ccall threadsafe  "Compression.h CompressWithHeader"
   c_CompressWithHeader   :: CMethod -> FunPtr INOUT_FUNC -> FunPtr INOUT_FUNC -> IO Int

-- |Decompress data compressed with c_CompressWithHeader (method name is read from compressed stream)
foreign import ccall threadsafe  "Compression.h DecompressWithHeader"
   c_DecompressWithHeader :: FunPtr INOUT_FUNC -> FunPtr INOUT_FUNC -> IO Int


----------------------------------------------------------------------------------------------------
----- High-level compression/decompression routines working with memory buffers --------------------
----------------------------------------------------------------------------------------------------

-- |Compress memory block
foreign import ccall threadsafe  "Compression.h CompressMem"
   c_CompressMem             :: CMethod -> Ptr CChar -> Int -> Ptr CChar -> Int -> IO Int

-- |Decompress memory block
foreign import ccall threadsafe  "Compression.h DecompressMem"
   c_DecompressMem           :: CMethod -> Ptr CChar -> Int -> Ptr CChar -> Int -> IO Int

-- |Compress memory block and save method name in compressed output
foreign import ccall threadsafe  "Compression.h CompressMemWithHeader"
   c_CompressMemWithHeader   :: CMethod -> Ptr CChar -> Int -> Ptr CChar -> Int -> IO Int

-- |Decompress memory block compressed with c_CompressMemWithHeader (method name is read from compressed data)
foreign import ccall threadsafe  "Compression.h DecompressMemWithHeader"
   c_DecompressMemWithHeader :: Ptr CChar -> Int -> Ptr CChar -> Int -> IO Int


----------------------------------------------------------------------------------------------------
----- Functions to query/modify compression methods ------------------------------------------------
----------------------------------------------------------------------------------------------------

-- |Returns canonical representation of compression method or error code
foreign import ccall unsafe  "Compression.h CanonizeCompressionMethod"
   c_CanonizeCompressionMethod   :: CMethod -> CMethod -> IO Int

-- |Returns memory used to compress/decompress, dictionary or block size of method given
foreign import ccall unsafe  "Compression.h GetCompressionMem"   c_GetCompressionMem   :: CMethod -> IO Int
foreign import ccall unsafe  "Compression.h GetDecompressionMem" c_GetDecompressionMem :: CMethod -> IO Int
foreign import ccall unsafe  "Compression.h GetDictionary"       c_GetDictionary       :: CMethod -> IO Int
foreign import ccall unsafe  "Compression.h GetBlockSize"        c_GetBlockSize        :: CMethod -> IO Int

-- |Set memory used to compress/decompress, dictionary or block size of method given
foreign import ccall unsafe  "Compression.h SetCompressionMem"   c_SetCompressionMem   :: CMethod -> Int -> CMethod -> IO Int
foreign import ccall unsafe  "Compression.h SetDecompressionMem" c_SetDecompressionMem :: CMethod -> Int -> CMethod -> IO Int
foreign import ccall unsafe  "Compression.h SetDictionary"       c_SetDictionary       :: CMethod -> Int -> CMethod -> IO Int
foreign import ccall unsafe  "Compression.h SetBlockSize"        c_SetBlockSize        :: CMethod -> Int -> CMethod -> IO Int

-- |Put upper limit to memory used to compress/decompress, dictionary or block size of method given
foreign import ccall unsafe  "Compression.h LimitCompressionMem"   c_LimitCompressionMem   :: CMethod -> Int -> CMethod -> IO Int
foreign import ccall unsafe  "Compression.h LimitDecompressionMem" c_LimitDecompressionMem :: CMethod -> Int -> CMethod -> IO Int
foreign import ccall unsafe  "Compression.h LimitDictionary"       c_LimitDictionary       :: CMethod -> Int -> CMethod -> IO Int
foreign import ccall unsafe  "Compression.h LimitBlockSize"        c_LimitBlockSize        :: CMethod -> Int -> CMethod -> IO Int


----------------------------------------------------------------------------------------------------
----- Direct calls to (de)compression rountines ----------------------------------------------------
----------------------------------------------------------------------------------------------------

-- |PPMD compression
foreign import ccall threadsafe  "PPMD/C_PPMD.h ppmd_compress"
   ppmd_compress   :: Int -> Int -> Int -> FunPtr INOUT_FUNC -> FunPtr INOUT_FUNC -> IO Int

-- |PPMD decompression
foreign import ccall threadsafe  "PPMD/C_PPMD.h ppmd_decompress"
   ppmd_decompress :: Int -> Int -> Int -> FunPtr INOUT_FUNC -> FunPtr INOUT_FUNC -> IO Int

-- |LZP compression
foreign import ccall threadsafe  "PPMD/C_LZP.h lzp_compress"
   lzp_compress    :: Int -> Int -> Int -> FunPtr INOUT_FUNC -> FunPtr INOUT_FUNC -> IO Int

-- |LZP decompression
foreign import ccall threadsafe  "PPMD/C_LZP.h lzp_decompress"
   lzp_decompress  :: Int -> Int -> Int -> FunPtr INOUT_FUNC -> FunPtr INOUT_FUNC -> IO Int

-- |LZMA compression
foreign import ccall threadsafe  "LZMA/C_LZMA.h lzma_compress"
   lzma_compress   :: Int -> Int -> Int -> Int -> Int -> Int -> Int -> FunPtr INOUT_FUNC -> FunPtr INOUT_FUNC -> IO Int

-- |LZMA decompression
foreign import ccall threadsafe  "LZMA/C_LZMA.h lzma_decompress"
   lzma_decompress :: Int -> Int -> Int -> Int -> Int -> Int -> Int -> FunPtr INOUT_FUNC -> FunPtr INOUT_FUNC -> IO Int

-- |BCJ-x86 preprocessing - encoder
foreign import ccall threadsafe  "LZMA/C_BCJ.h bcj_x86_compress"
   bcj_x86_compress    :: FunPtr INOUT_FUNC -> FunPtr INOUT_FUNC -> IO Int

-- |BCJ-x86 preprocessing - decoder
foreign import ccall threadsafe  "LZMA/C_BCJ.h bcj_x86_decompress"
   bcj_x86_decompress  :: FunPtr INOUT_FUNC -> FunPtr INOUT_FUNC -> IO Int

-- |GRZip compression
foreign import ccall threadsafe  "GRZip/C_GRZip.h grzip_compress"
   grzip_compress  :: Int -> Int -> Int -> Int -> Int -> Int -> Int -> Int -> FunPtr INOUT_FUNC -> FunPtr INOUT_FUNC -> IO Int

-- |GRZip decompression
foreign import ccall threadsafe  "GRZip/C_GRZip.h grzip_decompress"
   grzip_decompress:: Int -> Int -> Int -> Int -> Int -> Int -> Int -> Int -> FunPtr INOUT_FUNC -> FunPtr INOUT_FUNC -> IO Int


-- |Callback function type
type INOUT_FUNC  =  Ptr CChar -> Int -> IO Int
foreign import ccall threadsafe "wrapper"
   mkInOutFunc :: INOUT_FUNC -> IO (FunPtr INOUT_FUNC)

-- |Compression method representation
type Method = String

-- |Compression method representation in C
type CMethod = CString
