---------------------------------------------------------------------------------------------------
----  /     . ----------------------------------
---------------------------------------------------------------------------------------------------
module Errors
  ( ErrorTypes(..)
  , registerError
  , registerWarning
  , handleCtrlBreak
  , programTerminated
  ) where

import Prelude hiding (catch)
#if defined(FREEARC_WIN) && defined(__GLASGOW_HASKELL__)
import Control.Concurrent
import Control.Exception
import GHC.ConsoleHandler
#endif
import Data.Char
import System.IO.Unsafe
import System.Exit

import Utils


-- |     
data ErrorTypes = GENERAL_ERROR               String
                | CMDLINE_NO_COMMAND          [String]
                | CMDLINE_NO_ARCSPEC          [String]
                | CMDLINE_NO_FILENAMES        [String]
                | UNKNOWN_CMD                 String [String]
                | CMDLINE_UNKNOWN_OPTION      String
                | CMDLINE_AMBIGUOUS_OPTION    String [String]
                | CMDLINE_BAD_OPTION_FORMAT   String
                | INVALID_OPTION_VALUE        String String [String]
                | CANT_READ_DIRECTORY         String
                | CANT_GET_FILEINFO           String
                | CANT_OPEN_FILE              String
                | BAD_CRC                     String
                | TERMINATED
                | NOFILES
                | SKIPPED_FAKE_FILES          Int
                | BROKEN_ARCHIVE              String
                | INTERNAL_ERROR              String

-- |      
registerError err = do
  fail$ str2terminal$ "ERROR: "++errormsg err

-- |      
registerWarning warn = do
  putStr$ "\nWARNING: " ++ (str2terminal$ errormsg warn) ++ "  "

--foreign import "&errCounter" :: Ptr Int
{-
data SqliteException = SqliteException Int String
  deriving (Typeable)

catchSqlite :: IO a -> (SqliteException -> IO a) -> IO a
catchSqlite = catchDyn

throwSqlite :: SqliteException -> a
throwSqlite = throwDyn
-}

---------------------------------------------------------------------------------------------------
----   Ctrl-Break, Close  ..   -------------------------------
---------------------------------------------------------------------------------------------------

-- |    ,    Ctrl-Break
programTerminated = unsafePerformIO (ref False)

#if defined(FREEARC_WIN) && defined(__GLASGOW_HASKELL__)

-- | Ctrl-Break      
-- (,     ),   . , 
--  ,   
handleCtrlBreak handler action = do
  myThread <- myThreadId
  let onBreak event = do
        putStr " ^Break!"
        programTerminated =: True
        handler
        killThread myThread
  --         
  bracket (installHandler$ Catch onBreak) (installHandler) $  \oldHandler -> do
    action

#else

handleCtrlBreak handler action = action

#endif


---------------------------------------------------------------------------------------------------
----      .     --------------
---------------------------------------------------------------------------------------------------

errormsg (GENERAL_ERROR str) =
  str

errormsg (UNKNOWN_CMD cmd known_cmds) =
  "command "++quote cmd++" is unknown. Supported commands are " ++ join_with "," known_cmds

errormsg (CMDLINE_UNKNOWN_OPTION option) =
  "unknown option " ++ quote option

errormsg (CMDLINE_AMBIGUOUS_OPTION option variants) =
  "ambiguous option " ++ quote option ++ ": is that "++enumerate "or" variants++"?"

errormsg (CMDLINE_BAD_OPTION_FORMAT option) =
  "option " ++ quote option ++ " have illegal format"

errormsg (INVALID_OPTION_VALUE fullname shortname valid_values) =
  fullname++" option must be one of: "++ enumerate "or" (map (('-':shortname)++) valid_values)

errormsg (CMDLINE_NO_COMMAND args) =
  "no command name in command: " ++ quote (unwords args)

errormsg (CMDLINE_NO_ARCSPEC args) =
  "no archive name in command: " ++ quote (unwords args)

errormsg (CMDLINE_NO_FILENAMES args) =
  "no filenames in command: " ++ quote (unwords args)

errormsg (CANT_READ_DIRECTORY dir) =
  "error while reading directory " ++ quote dir

errormsg (CANT_GET_FILEINFO filename) =
  "can't get info about file " ++ quote filename

errormsg (CANT_OPEN_FILE filename) =
  "can't open file " ++ quote filename

errormsg (BAD_CRC filename) =
  "CRC error in " ++ filename

errormsg (TERMINATED) =
  "Program terminated!"

errormsg (NOFILES) =
  "No files, erasing empty archive"

errormsg (SKIPPED_FAKE_FILES n) =
  "skipped "++show n++" fake files"

errormsg (BROKEN_ARCHIVE msg) =
  "broken archive: "++msg

errormsg (INTERNAL_ERROR msg) =
  "FreeArc internal error: "++msg

-- |  
enumerate s list  =  join_with2 ", " (" "++s++" ") (map quote list)

{-# NOINLINE registerError #-}
{-# NOINLINE registerWarning #-}
{-# NOINLINE errormsg #-}
