----------------------------------------------------------------------------------------------------
----      ,    .                           ----
----------------------------------------------------------------------------------------------------
module FileInfo where

import Prelude hiding (catch)
import Control.Exception
import Control.Monad
import Data.Char
import Data.Int
import Data.IORef
import Data.List
import Data.Maybe
import Foreign.C
import System.Directory
import System.Posix.Internals

import Utils
import Process
import Files
import Errors
#ifdef FREEARC_PACKED_STRINGS
import FastString
#endif
#if defined(FREEARC_WIN)
import Win32Files
#endif


----------------------------------------------------------------------------------------------------
----     ----------------------------------------------------------
----------------------------------------------------------------------------------------------------

-- |         
--   ,       
data PackedFilePath = PackedFilePath
  { fpPackedDirectory   :: !MyPackedString     --  
  , fpPackedBasename    :: !MyPackedString     --    ,   
  , fpLCExtension       :: !String             -- ,    
  , fpHash              ::  Int32              --     ( ,           )
  , fpParent            ::  PackedFilePath     --  PackedFilePath   ( ,        -     error "...")
  }

instance Eq PackedFilePath where
  (==)  =  map2eq$ map3 (fpHash,fpPackedBasename,fpPackedDirectory)

#ifdef FREEARC_PACKED_STRINGS
--        2.5 !
type MyPackedString = FastString
myPackStr           = mkFastString
myUnpackStr         = unpackFS
#else
type MyPackedString = String
myPackStr           = id
myUnpackStr         = id
#endif

fpDirectory  =  myUnpackStr.fpPackedDirectory
fpBasename   =  myUnpackStr.fpPackedBasename

-- | :   ,    
fpFullname fp  = joinDirFilename (fpDirectory fp) (fpBasename fp)

-- |     
packFilePath parent fullname  =  packFilePath2 parent dir name
  where (dir,name) = splitDirFilename fullname

-- |          
packFilePath2       parent dir        name  =  packFilePathPacked2 parent (myPackStr dir) name
packFilePathPacked2 parent packed_dir name  =  packFilePathPacked3 parent packed_dir name (filenameLower$ getFileSuffix name)

-- |     ,      
packFilePath3 parent dir name lcext              =  packFilePathPacked3 parent (myPackStr dir) name lcext
packFilePathPacked3 parent packed_dir name lcext =
  PackedFilePath { fpPackedDirectory    =  packed_dir
                 , fpPackedBasename     =  myPackStr name
                 , fpLCExtension        =  lcext
                 , fpHash               =  filenameHash (fpHash parent) name
                 , fpParent             =  parent
                 }

-- |       
packParentDirPath dir  =
  PackedFilePath { fpPackedDirectory    =  myPackStr ""   --     ,
                 , fpPackedBasename     =  myPackStr dir  --      Basename
                 , fpLCExtension        =  ""
                 , fpHash               =  filenameHash 0 (filter (not.isPathSeparator) dir)
                 , fpParent             =  error "No parent for Parent Directory"
                 }

-- |     (  !).
--      `dirhash` -   ,  ,
--  `basename` -     
filenameHash {-dirhash basename-}  =  foldl (\h c -> h*37+i(ord c))



----------------------------------------------------------------------------------------------------
----       -------------------------------------------
----------------------------------------------------------------------------------------------------

-- |     `filespec`.
--  "*", "*.ext"     ,  
match_FP getName filespec =
  if filespec==reANY_FILE  then const True  else
    case (splitFilename3 filespec) of
      ("", "*", ext) -> match  (filenameLower ext)      . fpLCExtension
      ("", _,   _  ) -> match  (filenameLower filespec) . filenameLower . getName
      _              -> match  (filenameLower filespec) . filenameLower . fpFullname

-- |     `filepath`     `filespecs`?
match_filespecs getName filespecs {-filepath-}  =  any_function (map (match_FP getName) filespecs)

-- |,     
reANY_FILE = "*"



----------------------------------------------------------------------------------------------------
----    ----------------------------------------------------------------------------
----------------------------------------------------------------------------------------------------

type FileSize = Integer    --         /  
type FileTime = CTime      --      // 

-- |,       
data FileInfo = FileInfo
  { fiFilteredName :: !PackedFilePath    --  ,      
  , fiDiskName     :: !PackedFilePath    -- ""   -  /   
  , fiStoredName   :: !PackedFilePath    -- ""   -    
  , fiSize         :: !FileSize          --   (0  )
  , fiTime         :: !FileTime          -- /  
  , fiIsDir        :: !Bool              --  ?
  }

-- | FileInfo     
diskName = fpFullname.fiDiskName

-- |  FileInfo     
createParentDirFileInfo fiFilteredName fiDiskName fiStoredName =
  FileInfo { fiFilteredName  =  packParentDirPath fiFilteredName
           , fiDiskName      =  packParentDirPath fiDiskName
           , fiStoredName    =  packParentDirPath fiStoredName
           , fiSize          =  0
           , fiTime          =  0
           , fiIsDir         =  True
           }

-- |       ( ,    )
rereadFileInfo fi file = do
  getFileInfo (fiFilteredName fi) (fiDiskName fi) (fiStoredName fi)

-- |  FileInfo     
getFileInfo fiFilteredName fiDiskName fiStoredName  =
    let filename = fpFullname fiDiskName in do
    fileWithStatus "getFileInfo" filename $ \p_stat -> do
      fiSize   <-  stat_size  p_stat
      fiTime   <-  stat_mtime p_stat
      fiIsDir  <-  stat_mode  p_stat  >>==  s_isdir
      return$ Just$ FileInfo fiFilteredName fiDiskName fiStoredName fiSize fiTime fiIsDir
  `catch`
    \e -> do registerWarning$ CANT_GET_FILEINFO filename
             return Nothing  --      stat  Nothing

-- |Restore date/time/attrs saved in FileInfo structure
setFileDateTimeAttr filename fileinfo  =  setFileDateTime filename (fiTime fileinfo)

{-# NOINLINE getFileInfo #-}


----------------------------------------------------------------------------------------------------
----     -------------------------------------------------------------------------
----------------------------------------------------------------------------------------------------

#if !defined(FREEARC_WIN)

-- |A version of getDirectoryContents that is non-fatal if the directory doesn't exist.
softGetDirectoryContents dir = do
  catchJust ioErrors
    (getDirectoryContents (dir ||| "."))
    (\e -> do registerWarning$ CANT_READ_DIRECTORY dir; return [])

-- |      ( "."  ".."),    `parent`
getDirectoryContents_FileInfo basedirs parent{-  FileInfo-} = do
  let filteredDirName = fpFullname (fiFilteredName parent)   --   
      diskDirName     = fpFullname (fiDiskName parent)       --
      storedDirName   = fpFullname (fiStoredName parent)     --
      readFileInfo = getFileInfoOptimized basedirs parent (myPackStr filteredDirName) (myPackStr diskDirName) (myPackStr storedDirName)
  softGetDirectoryContents diskDirName   --     
    >>== filter exclude_special_names    --    "."  ".."
    >>=  mapMaybeM readFileInfo          --      FileInfo     ,    `stat`

-- |         .
getFileInfoOptimized (arc_basedir,disk_basedir) parent filteredDirName diskDirName storedDirName name = do
  let filteredName  =  packFilePathPacked3 (fiFilteredName parent) filteredDirName name lcext
      diskName      =  packFilePathPacked3 (fiDiskName     parent) diskDirName     name lcext
      storedName    =  packFilePathPacked3 (fiStoredName   parent) storedDirName   name lcext
      lcext         =  filenameLower$ getFileSuffix name
  getFileInfo (if disk_basedir>"" then filteredName else diskName)
              diskName
              (if disk_basedir>"" || arc_basedir>"" then storedName else diskName)

{-# INLINE getFileInfoOptimized #-}

#else

getDirectoryContents_FileInfo (arc_basedir,disk_basedir) parent{-  FileInfo-} = do
  let filteredDirName = fpFullname (fiFilteredName parent)   --   
      diskDirName     = fpFullname (fiDiskName parent)       --
      storedDirName   = fpFullname (fiStoredName parent)     --
  handle (\e -> do registerWarning$ CANT_READ_DIRECTORY diskDirName; return []) $ do
    withList $ \list -> do
      wfindfiles (if diskDirName>"" then diskDirName++"/*" else "*") $ \find -> do
        name <- w_find_name find
        when (exclude_special_names name) $ do
          fiSize  <- w_find_size       find
          fiTime  <- w_find_time_write find
          fiIsDir <- w_find_isDir      find
          let filteredName  =  packFilePathPacked3 (fiFilteredName parent) (myPackStr filteredDirName) name lcext
              diskName      =  packFilePathPacked3 (fiDiskName     parent) (myPackStr diskDirName)     name lcext
              storedName    =  packFilePathPacked3 (fiStoredName   parent) (myPackStr storedDirName)   name lcext
              lcext         =  filenameLower$ getFileSuffix name
          let fiFilteredName = if disk_basedir>"" then filteredName else diskName
              fiDiskName     = diskName
              fiStoredName   = if disk_basedir>"" || arc_basedir>"" then storedName else diskName
          list <<= FileInfo fiFilteredName fiDiskName fiStoredName fiSize fiTime fiIsDir

#endif


-- |    `dir`,  `accept_f`     `process_f`.
--  scan_subdirs==True -        
findFiles_FileInfo scan_subdirs accept_f process_f basedirs dir = do
  if scan_subdirs  then recursiveM processDir dir  else do processDir dir; return ()
  where processDir dir = do
          dirContents  <-  getDirectoryContents_FileInfo basedirs dir
          let files     =  filter accept_f dirContents   --  ,  
              subdirs   =  filter fiIsDir  dirContents   --  
          unless (null files) (process_f files)          --  ,    
          return subdirs                                 --      

{-# NOINLINE getDirectoryContents_FileInfo #-}
{-# NOINLINE findFiles_FileInfo #-}


----------------------------------------------------------------------------------------------------
----    ,    ----------------------------------
----------------------------------------------------------------------------------------------------

-- | []  ,   `filespec`,    
find_files scan_subdirs filespec  =  find_and_filter_files scan_subdirs (Just False) (const True) "" "" [filespec]

-- | (, )  ,   `filespecs`,
--    `filter_f`,    
find_and_filter_files scan_subdirs include_dirs filter_f arc_basedir disk_basedir filespecs = do
  concat ==<< withList   --    ,    
    (\list -> find_filter_and_process_files scan_subdirs include_dirs filter_f (list<<=) arc_basedir disk_basedir filespecs)

-- | (, )  ,   `filespecs`,
--    `filter_f`,          
find_and_filter_files_PROCESS scan_subdirs include_dirs filter_f arc_basedir disk_basedir filespecs pipe = do
  find_filter_and_process_files scan_subdirs include_dirs filter_f (sendP pipe) arc_basedir disk_basedir filespecs
  sendP pipe []  --  " -  !" :)


-- | []  ,   `filespecs`    `filter_f`,
--      ,    ,  `process_f`
find_filter_and_process_files scan_subdirs include_dirs filter_f process_f arc_basedir disk_basedir {-filespecs-} =

  --     ,       
  mapM_ find_files_in_one_dir . sort_and_groupOn (filenameLower.directoryOf)  where

    --   ,    
    find_files_in_one_dir filespecs
      = findFiles_FileInfo do_scan_subdirs accept_f process_f (arc_basedir,disk_basedir) dir

      where dirname = directoryOf (head filespecs)            --     
            masks   = map basenameOf filespecs                --     
            dir     = createParentDirFileInfo                   --   FileInfo   :
                        dirname                                 --       
                        (joinDirFilename disk_basedir dirname)  --      
                        (joinDirFilename arc_basedir dirname)   --      
            --      "-r"       "dir/"
            do_scan_subdirs  =  scan_subdirs  ||  masks `contains` ""
            --     /,       "dir/"  "*"  "dir/*"
            include_all      =  masks `contains_one_of` ["", reANY_FILE]
            -- ,          :
            --         --[no]dirs, by default -   "* -r" || "dir/"
            --        `filter_f`    
            accept_f fi | fiIsDir fi  =  include_dirs `defaultVal` (do_scan_subdirs && include_all)
                        | otherwise   =  filter_f fi && (include_all || match_filespecs fpBasename masks (fiFilteredName fi))


{-# NOINLINE find_files #-}
{-# NOINLINE find_and_filter_files #-}
{-# NOINLINE find_and_filter_files_PROCESS #-}
{-# NOINLINE find_filter_and_process_files #-}
