----------------------------------------------------------------------------------------------------
----    ,     , /.                     ----
----------------------------------------------------------------------------------------------------
module Files where

import Prelude hiding (catch)
import Control.Concurrent
import Control.Concurrent.MVar
import Control.Exception
import Control.Monad
import Data.Char
import Foreign
import Foreign.C
import Foreign.Marshal.Alloc
import System.Directory
import System.Posix.Internals
import System.Posix.Types
import System.IO
import System.IO.Unsafe
import System.Locale
import System.Time

import Utils
import Errors
#if defined(FREEARC_WIN)
import Win32Files
#endif

-- |  
aBUFFER_SIZE = 64*kb

----------------------------------------------------------------------------------------------------
---- Filename manipulation -   GHC/compiler/main/DriverUtil.hs ---------------------------
----------------------------------------------------------------------------------------------------

splitFilenameSuffix :: String -> (String,String)
splitFilenameSuffix f = split_prefix_suffix '.' f

getFilePrefix :: String -> String
getFilePrefix f = remove_suffix '.' f

getFileSuffix :: String -> String
getFileSuffix f = remove_prefix '.' f

joinFilename :: String -> String -> String
joinFilename prefix suffix =
  case (suffix) of
    ""   ->  prefix
    _    ->  prefix ++ '.':suffix

-- "foo/bar/xyzzy.ext" -> ("foo/bar", "xyzzy.ext")
splitDirFilename :: String -> (String,String)
splitDirFilename str = split_longest_prefix str isPathSeparator

-- "foo/bar/xyzzy.ext" -> ("foo/bar", "xyzzy", "ext")
splitFilename3 :: String -> (String,String,String)
splitFilename3 str
   = let (dir, rest) = split_longest_prefix str isPathSeparator
         (name, ext) = splitFilenameSuffix rest
     in  (dir, name, ext)

-- "foo/bar" "xyzzy" "ext" -> "foo/bar/xyzzy.ext"
joinFilename3 :: String -> String -> String -> FilePath
joinFilename3 dir base ext  =  joinDirFilename dir (joinFilename base ext)

remove_suffix :: Char -> String -> String
remove_suffix c s
  | null pre  = s
  | otherwise = reverse (tail pre)
  where (suf,pre) = break (==c) (reverse s)

remove_prefix :: Char -> String -> String
remove_prefix c s
  | null pre  = ""
  | otherwise = reverse suf
  where (suf,pre) = break (==c) (reverse s)

split_prefix_suffix :: Char -> String -> (String,String)
split_prefix_suffix c s
  | null pre  = (s,"")
  | otherwise = (reverse (tail pre), reverse suf)
  where (suf,pre) = break (==c) (reverse s)

drop_longest_prefix :: String -> (Char -> Bool) -> String
drop_longest_prefix s pred = reverse suf
  where (suf,_pre) = break pred (reverse s)

take_longest_prefix :: String -> (Char -> Bool) -> String
take_longest_prefix s pred = reverse pre
  where (_suf,pre) = break pred (reverse s)

-- split a string at the last character where 'pred' is True,
-- returning a pair of strings. The first component holds the string
-- up (but not including) the last character for which 'pred' returned
-- True, the second whatever comes after (but also not including the
-- last character).
--
-- If 'pred' returns False for all characters in the string, the original
-- string is returned in the second component (and the first one is just
-- empty).
split_longest_prefix :: String -> (Char -> Bool) -> (String,String)
split_longest_prefix s pred
  = case pre of
        []      -> ([], reverse suf)
        (_:pre) -> (reverse pre, reverse suf)
  where (suf,pre) = break pred (reverse s)

replaceFilenameSuffix :: FilePath -> String -> FilePath
replaceFilenameSuffix s suf = (remove_suffix '.' s) ++ suf

updateFilenameBasenamePrefix :: FilePath -> (String -> String) -> FilePath
updateFilenameBasenamePrefix path update_func  =  joinFilename3 dir (update_func base) ext
  where (dir,base,ext) = splitFilename3 path

-- directoryOf strips the filename off the input string, returning
-- the directory.
directoryOf :: FilePath -> String
directoryOf =  fst . splitDirFilename

-- |   
basenameOf :: FilePath -> String
basenameOf str  =  drop_longest_prefix str isPathSeparator

joinDirFilename :: String -> FilePath -> FilePath
joinDirFilename dir name =
  case dir of
    ""   ->  name
    "."  ->  name
    _    ->  case name of
               ""   ->  dir
               _    ->  dir ++ pathSeparator:name

replaceFilenameDirectory :: FilePath -> String -> FilePath
replaceFilenameDirectory s dir
 = joinDirFilename dir (drop_longest_prefix s isPathSeparator)

remove_spaces :: String -> String
remove_spaces = reverse . dropWhile isSpace . reverse . dropWhile isSpace

escapeSpaces :: String -> String
escapeSpaces = foldr (\c s -> if isSpace c then '\\':c:s else c:s) ""

pathSeparator = '/'

isPathSeparator :: Char -> Bool
isPathSeparator ch =
#if defined(FREEARC_WIN)
  ch=='/' || ch=='\\'
#else
  ch=='/'
#endif

-- |True,     
havePath  =  any isPathSeparator

-- |  '\'  '/'
translatePath = map (\c -> if isPathSeparator c  then pathSeparator  else c)

stripPathSeparatorAtTheEnd s =
  if  s>""  &&  isPathSeparator (last s)
  then init s
  else s

-- |True,  file    `dir`,    ,    
dir `isParentDirOf` file =
  case (start_from dir file) of
    Just ""    -> True
    Just (x:_) -> isPathSeparator x
    Nothing    -> False


#if defined(FREEARC_WIN)
-- | case-insensitive  
filenameLower = strLower
#else
-- | case-sensitive  
filenameLower = id
#endif

-- |Return False for special filenames like "." and "..", used to filtering results of getDirContents
exclude_special_names s  =  (s/=".")  &&  (s/="..")


----------------------------------------------------------------------------------------------------
----       ---------------------------------------------------
----------------------------------------------------------------------------------------------------

-- |Create a hierarchy of directories
createDirectoryHierarchy :: FilePath -> IO ()
createDirectoryHierarchy dir = do
  when (dir/= "" && exclude_special_names dir) $ do
    b <- dirExist dir
    when (not b) $ do
      createDirectoryHierarchy (directoryOf dir)
      dirCreate dir

-- |      
buildPathTo filename  =  createDirectoryHierarchy (directoryOf filename)

-- |      
temparc_prefix = "temparc"
temparc_suffix = ".tmp"

-- | `action`        
tempfile_wrapper filename action = do
  let tempname = temparc_prefix++temparc_suffix
      run_action_with name  =  action name  `catch`  (\e -> do fileRemove name; throwIO e)
  old_file <- fileExist filename
  if old_file
    then do x <- run_action_with tempname
            fileRemove filename
            fileRename tempname filename
            return x
    else do run_action_with filename

-- | datetime,      .  1  1970 .
aMINIMAL_POSSIBLE_DATETIME = 0 :: CTime

-- |Get file's date/time
getFileDateTime filename  =  fileWithStatus "getFileDateTime" filename stat_mtime

-- |Set file's date/time
setFileDateTime filename datetime  =  withCFilePath filename (`c_SetFileDateTime` datetime)

-- | CTime  ClockTime.      ClockTime  GHC!!!
convert_CTime_to_ClockTime ctime = TOD (realToInteger ctime) 0
  where realToInteger = round . realToFrac :: Real a => a -> Integer

-- |  
showtime format t = formatCalendarTime defaultTimeLocale format (unsafePerformIO (toCalendarTime t))

-- | ,   
tryFileRemove filename  =  whenM (fileExist filename) (fileRemove filename)

-- | -           ""
findConfigFile filename = do
  dir <- getProgramDir
  let default_file = joinDirFilename dir filename
  exist <- fileExist default_file
  return$  if exist  then default_file  else ""


#if defined(__GLASGOW_HASKELL__)

-- |,     
getProgramDir :: IO String
getProgramDir =
  alloca $ \ p_argc -> do
  alloca $ \ p_argv -> do
    getProgArgv p_argc p_argv
    argv <- peek p_argv
    s <- peekElemOff argv 0 >>= c_ExpandProgname >>= peekCString
    return (directoryOf s)

foreign import ccall unsafe "getProgArgv"
  getProgArgv :: Ptr CInt -> Ptr (Ptr CString) -> IO ()

foreign import ccall unsafe "Environment.h ExpandProgname"
  c_ExpandProgname :: CString -> IO CString

#else

getProgramDir :: IO String
getProgramDir = return "."

#endif


----------------------------------------------------------------------------------------------------
----     ------------------------------------------------------------------
----------------------------------------------------------------------------------------------------

--withMVar  mvar action     =  bracket (takeMVar mvar) (putMVar mvar) action
liftMVar1  action mvar     =  withMVar mvar action
liftMVar2  action mvar x   =  withMVar mvar (\a -> action a x)
liftMVar3  action mvar x y =  withMVar mvar (\a -> action a x y)
returnMVar action          =  action >>= newMVar

-- | ,   MVar          
type Archive         = MVar File
archiveCreate        = returnMVar . fileCreate
archiveOpen          = returnMVar . fileOpen
archiveGetPos        = liftMVar1 fileGetPos
archiveGetSize       = liftMVar1 fileGetSize
archiveSeek          = liftMVar2 fileSeek
archiveRead          = liftMVar2 fileRead
archiveReadBuf       = liftMVar3 fileReadBuf
archiveWrite         = liftMVar2 fileWrite
archiveWriteBuf      = liftMVar3 fileWriteBuf
archiveClose         = liftMVar1 fileClose

-- |             
archiveCopyData srcarc pos size dstarc = do
  withMVar srcarc $ \srcfile ->
    withMVar dstarc $ \dstfile -> do
      restorePos <- fileGetPos srcfile
      fileSeek      srcfile pos
      fileCopyBytes srcfile size dstfile
      fileSeek      srcfile restorePos


#if defined(FREEARC_WIN)

type File            = FD
type CFilePath       = CWFilePath
withCFilePath        = withCWFilePath
fileCreate name      = wopen name (write_flags .|. o_BINARY .|. o_TRUNC) 0o666
fileOpen   name      = wopen name (read_flags .|. o_BINARY) 0o666
fileGetPos           = wtell
fileGetSize          = wfilelength
fileSeek   file pos  = wseek file pos sEEK_SET
fileRead   file size = allocaBytes size $ \buf -> do fileReadBuf file buf size; peekCStringLen (buf,size)
fileWrite  file str  = withCStringLen str $ \(buf,size) -> fileWriteBuf file buf size
fileReadBuf          = wread
fileWriteBuf         = wwrite
fileGet    name      = bracket (fileOpen name) fileClose (\file -> fileGetSize file >>= fileRead file)
fileClose            = wclose
fileExist            = wDoesFileExist
fileRemove           = wunlink
fileRename           = wrename
fileWithStatus       = wWithFileStatus
stat_mode            = wst_mode
stat_size            = wst_size
stat_mtime           = wst_mtime
dirCreate            = wmkdir
dirExist             = wDoesDirectoryExist
dirRemove            = wrmdir

#else

type File            = Handle
type CFilePath       = CString
withCFilePath        = withCString
fileCreate name      = openBinaryFile name WriteMode
fileOpen   name      = openBinaryFile name ReadMode
fileGetPos           = hTell
fileGetSize          = hFileSize
fileSeek   file      = hSeek file AbsoluteSeek
fileRead   file size = replicateM size (hGetChar file)
fileWrite            = hPutStr
fileReadBuf          = hGetBuf
fileWriteBuf         = hPutBuf
fileGet    name      = hGetContents =<< fileOpen name
fileClose            = hClose
fileExist            = doesFileExist
fileRemove           = removeFile
fileRename           = renameFile
fileSetSize          = hSetFileSize
fileWithStatus       = withFileStatus
stat_mode            = st_mode
stat_size            = st_size  .>>== i
stat_mtime           = st_mtime
dirCreate            = createDirectory
dirExist             = doesDirectoryExist
dirRemove            = removeDirectory

#endif

filePut name str  =  bracket (fileCreate name) fileClose (`fileWrite` str)

-- |         
fileCopyBytes srcfile size dstfile = do
  allocaBytes aBUFFER_SIZE $ \buf -> do  --  `alloca`,       
    doChunks size aBUFFER_SIZE $ \bytes -> do
      fileReadBuf  srcfile buf bytes
      fileWriteBuf dstfile buf bytes

-- | Nothing     ,     
tryOpen filename = catchJust ioErrors
                     (fileOpen filename >>== Just)
                     (\e -> do registerWarning$ CANT_OPEN_FILE filename; return Nothing)

----------------------------------------------------------------------------------------------------
----- External functions ---------------------------------------------------------------------------
----------------------------------------------------------------------------------------------------

-- |Set date/time/attrs for a file
foreign import ccall unsafe "Environment.h SetFileDateTime"
   c_SetFileDateTime :: CFilePath -> CTime -> IO ()

