---------------------------------------------------------------------------------------------------
----  :   , ,  ,           ----
----    .    IORef-,                            ----
----         .                          ----
---------------------------------------------------------------------------------------------------
module Utils where

import Control.Concurrent
import Control.Exception
import Control.Monad
import Data.Array
import Data.Char
import Data.Either
import Data.HashTable
import Data.IORef
import Data.List
import Data.Maybe
import Data.Word
import Debug.Trace
import Foreign.C.String
import Foreign.Marshal.Alloc
import Foreign.Marshal.Array
import Foreign.Marshal.Pool
import Foreign.Marshal.Utils
import Foreign.Ptr
import Text.Regex
import System.IO.Unsafe
#ifdef FREEARC_WIN
import System.Win32
#endif

---------------------------------------------------------------------------------------------------
----  define's ----------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------

#if !defined(FREEARC_WIN) && !defined(FREEARC_UNIX)
#error "You must define OS!"
#endif

#if !defined(FREEARC_INTEL_BYTE_ORDER) && !defined(FREEARC_MOTOROLA_BYTE_ORDER)
#error "You must define byte order!"
#endif


---------------------------------------------------------------------------------------------------
----  :) ------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------

-- | 4-    
#if defined(FREEARC_INTEL_BYTE_ORDER)
make4byte b0 b1 b2 b3 = b0+256*(b1+256*(b2+256*b3)) :: Word32
#else
make4byte b0 b1 b2 b3 = b3+256*(b2+256*(b1+256*b0)) :: Word32
#endif

-- |   - ,   ..
b :kb :mb :gb :_       = iterate (1024*) 1 :: [Int]
b_:kb_:mb_:gb_:tb_:_   = iterate (1024*) 1 :: [Integer]

-- | ,    .    ,
--    (b/k/f/...),    -  `default_specifier`.
--     ,     - `b`,     ,
--   ,   ,    `default_specifier`.
parseNumber num default_specifier =
  case (span isDigit$ strLower$ num++[default_specifier]) of
    (digits, 'b':_)  ->  (readI digits      , 'b')
    (digits, 'k':_)  ->  (readI digits * kb_, 'b')
    (digits, 'm':_)  ->  (readI digits * mb_, 'b')
    (digits, 'g':_)  ->  (readI digits * gb_, 'b')
    (digits, 't':_)  ->  (readI digits * tb_, 'b')
    (digits, '^':_)  ->  (2 ^ readI digits  , 'b')
    (digits,  c :_)  ->  (readI digits      ,  c )

-- |   : "512b", "32k", "8m"   . "24"  2^24 ( 7-zip)
parseMem mem  =  case (parseNumber mem '^')  of  (bytes, 'b')  ->  fromInteger bytes

readI = foldl f 0
  where f m c | isDigit c  =  fromIntegral (ord c - ord '0') + (m * 10)
              | otherwise  =  error ("Non-digit "++[c]++" in readI")

readInt :: String -> Int
readInt = fromInteger.readI

{-# NOINLINE parseNumber #-}
{-# NOINLINE parseMem #-}
{-# NOINLINE readI #-}
{-# NOINLINE readInt #-}


-- |      
infixl 9  .$
infixl 1  >>==, ==<<, =<<., .>>=, .>>, .>>==, ==<<.
a.$b         =  b a                --  $    
a>>==b       =  a >>= return.b     --  >>=,     c  
a==<<b       =  return.a =<< b     --  =<<,     c  
(a=<<.b) c   =  a =<< b c          --  =<<    mapM    
(a.>>=b) c   =  a c >>= b          --  >>=    mapM    
(a.>>b)  c   =  a c >> b           --  >>    mapM    
(a==<<.b) c  =  return.a =<< b c   --  ==<<    mapM    
(a.>>==b) c  =  a c >>= return.b   --  >>==    mapM    

-- |    
[] ||| b = b
a  ||| _ = a

-- |     .     ,    
[] &&& _ = []
_  &&& b = b

-- |  f       
unlessNull f xs  =  xs &&& f xs

-- |  f       
unlessNullM f []  =  return ()
unlessNullM f xs  =  f xs >> return ()

-- |  concatMap
concatMapM :: Monad io => (a -> io [b]) -> [a] -> io [b]
concatMapM f x  =  mapM f x  >>==  concat

-- | 
whenM cond action = do
  allow <- cond
  when allow
    action

-- | `action`  ,  `x`,    Nothing
whenJustM x action  =  x >>= maybe (return Nothing) action

-- | 
repeat_foreverM action = do
  action
  repeat_foreverM action

-- | ,   'while'   
repeat_whileM inp cond out = do
  x <- inp
  if (cond x)
    then do out x
            repeat_whileM inp cond out
    else return x

-- | ,  repeat-until  
repeat_untilM action = do
  done <- action
  when (not done) $ do
    repeat_untilM action

-- | ,     size
--       chunk 
doChunks size chunk action =
  case size of
    0 -> return ()
    _ -> do let n = minI size chunk
            action (fromIntegral n)
            doChunks (size-n) chunk action

-- | `action`  x,     ,   `action`,    
recursiveM action x  =  action x >>= mapM_ (recursiveM action)

-- | ,    `cond`,     
recursiveIfM cond action x  =  if cond  then recursiveM action x  else (action x >> return ())

-- |  `action`    `list` , 
--  ,  `action` -  ,  mapM.
--          `crit_f`    ,
--    .     
--   `list`
mapMConditional (init,map_f,sum_f,crit_f) action list = do
  let go []     ys summary = return (reverse ys, [])     --  -  
  let go (x:xs) ys summary = do
        y <- action x
        let summary  =  sum_f summary (map_f y)
        if (crit_f summary)
          then return (reverse$ y:ys, xs)                --   
          else go xs (y:ys) summary
  go list [] init

withThread thread  =  bracket (forkIO thread) killThread . const

-- |Must be in Foreign.C.String, but don't
pokeCStringLen (addr,len) inp  =  pokeArray addr (map castCharToCChar (take len inp))

{-# NOINLINE repeat_foreverM #-}
{-# NOINLINE repeat_whileM #-}
{-# NOINLINE repeat_untilM #-}
{-# NOINLINE mapMConditional #-}
{-# NOINLINE pokeCStringLen #-}


-- |     () 
filterM :: (Monad m) => (a -> m Bool) -> [a] -> m [a]
filterM p  =  go []
  where go accum []      =  return$ reverse accum
        go accum (x:xs)  =  p x  >>=  bool (go    accum  xs)
                                           (go (x:accum) xs)

-- |mapMaybe,    Monad
mapMaybeM :: Monad m => (a -> m (Maybe b)) -> [a] -> m [b]
mapMaybeM f  =  go []
  where go accum []      =  return$ reverse accum
        go accum (x:xs)  =  f x  >>=  maybe (      go    accum  xs)
                                            (\r -> go (r:accum) xs)

-- |   liftM,   - ,      :)
liftMaybe f (Just x)  =  Just (f x)
liftMaybe f Nothing   =  Nothing

-- | Nothing    
defaultVal = flip fromMaybe

-- |         
bool onFalse onTrue False  =  onFalse
bool onFalse onTrue True   =  onTrue

-- |  Left
isLeft (Left _) = True
isLeft _        = False

-- |    
trace2 s = trace (show s) s

{-
-- Cale Gibbard

A useful little higher order function. Some examples of use:

swing map :: forall a b. [a -> b] -> a -> [b]
swing any :: forall a. [a -> Bool] -> a -> Bool
swing foldr :: forall a b. b -> a -> [a -> b -> b] -> b
swing zipWith :: forall a b c. [a -> b -> c] -> a -> [b] -> [c]
swing find :: forall a. [a -> Bool] -> a -> Maybe (a -> Bool) -- applies each of the predicates to the given value, returning the first predicate which succeeds, if any
swing partition :: forall a. [a -> Bool] -> a -> ([a -> Bool], [a -> Bool])

-}

swing :: (((a -> b) -> b) -> c -> d) -> c -> a -> d
swing f = flip (f . flip ($))


-- |Map on functions instead of its' arguments!
map_functions []     x  =  []
map_functions (f:fs) x  =  f x : map_functions fs x

-- |,       True  ( ) . ,  swing all
all_functions []  = const True
all_functions [f] = f
all_functions fs  = and . map_functions fs

-- |,        True  ( ) . ,  swing any
any_function []  = const False
any_function [f] = f
any_function fs  = or . map_functions fs


#ifdef FREEARC_WIN

-- |Translate string from Unicode to OEM encoding
unicode2oem s =
  if all isAscii s
    then s
    else unsafePerformIO $ do
           withTString s $ \wstr -> do
             allocaBytes (length s + 1) $ \cstr -> do
               c_WideToOem wstr cstr
               peekCString cstr

foreign import stdcall unsafe "winuser.h CharToOemW"
  c_WideToOem :: LPCTSTR -> CString -> IO Bool


-- |Translate string from OEM encoding to Unicode
oem2unicode s =
  if all isAscii s
    then s
    else unsafePerformIO $ do
           withCString s $ \str -> do
             allocaBytes ((length s + 1) * 2) $ \wstr -> do
               c_OemToWide str wstr
               peekTString wstr

foreign import stdcall unsafe "winuser.h OemToCharW"
  c_OemToWide :: CString -> LPCTSTR -> IO Bool

-- |Translate string from ANSI encoding to Unicode
ansi2unicode s =
  if all isAscii s
    then s
    else unsafePerformIO $ do
           withCString s $ \str -> do
             allocaBytes ((length s + 1) * 2) $ \wstr -> do
               c_AnsiToOem str str
               c_OemToWide str wstr
               peekTString wstr

foreign import stdcall unsafe "winuser.h CharToOemA"
  c_AnsiToOem :: CString -> CString -> IO Bool


-- The following functions implementations assume that filenames are internally represented in Unicode
-- (see definitions of file* functions in File.hs, which uses Unicode filenames under Windows
--  and getDirectoryContents_FileInfo in FileInfo.hs, which returns Unicode filenames under Windows)

-- |Translate filename from internal to terminal encoding
str2terminal = unicode2oem
-- |Translate filename from filelist to internal encoding
filelist2str = oem2unicode
-- |Translate filename from cmdline to internal encoding
cmdline2str  = ansi2unicode

#else

-- The following functions implementations assume that internal and external representations of filenames are the same
-- |Translate filename from internal to terminal encoding
str2terminal = id
-- |Translate filename from filelist to internal encoding
filelist2str = id
-- |Translate filename from cmdline to internal encoding
cmdline2str  = id

#endif



---------------------------------------------------------------------------------------------------
----    ------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------

-- |   ,   
split :: (Eq a) => a -> [a] -> [[a]]
split c s =
  let (chunk, rest) = break (==c) s
  in case rest of  []     -> [chunk]
                   _:rest -> chunk : split c rest

-- |       : "one, two, three"
join_with :: [a] -> [[a]] -> [a]
join_with x xs =  concat$ intersperse x xs

-- |     ,    :
-- join_with2 ["one","two","three","four"]  ", "  " and "  -->  "one, two, three and four"
join_with2 :: [a] -> [a] -> [[a]] -> [a]
join_with2 a b []    =  []
join_with2 a b [x]   =  x
join_with2 a b list  =  join_with a (init list) ++ b ++ last list

-- | x  s1  s2,    
between s1 x [] = s1
between [] x s2 = s2
between s1 x s2 = s1++x++s2

-- |    
quote :: String -> String
quote str  =  "\"" ++ str ++ "\""

contains = flip elem

-- |,  `s`        `set`
s `contains_one_of` set  =  any (`elem` set) s

-- | n'  (  0)   `xs`  `x`
replaceAt n xs x  =  take n xs ++ x : drop (n+1) xs

-- |      -    ,  Nothing
start_from (x:xs) (y:ys) | x==y  =  start_from xs ys
start_from [] str                =  Just str
start_from _  _                  =  Nothing

-- |   `substr`   `str`
tryToSkip substr str  =  (start_from substr str) `defaultVal` str

-- |   `s`   `from`  `to`
replaceAll from to s =
  case (start_from from s, s) of
    (Just remainder, _   )  ->  to ++ replaceAll from to remainder
    (Nothing       , c:cs)  ->  c : replaceAll from to cs
    (Nothing       , []  )  ->  []

-- |   `s`  `from`  `to`
replaceAtEnd from to s =
  case start_from (reverse from) (reverse s) of
    Just remainder  -> reverse remainder ++ to
    Nothing         -> s

-- |  /,        - 
right_fill  c n s  =  s ++ replicate (n-length s) c
left_fill   c n s  =  replicate (n-length s) c ++ s
left_justify       =  right_fill ' '
right_justify      =  left_fill  ' '

-- |    
strLower = map toLower

{-# NOINLINE replaceAll #-}
{-# NOINLINE replaceAtEnd #-}



---------------------------------------------------------------------------------------------------
----    ------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------

-- |Sort list by function result (use Schwarznegian transform)
sortOn  f  =  map snd . sortOn' fst . map (keyval f)

-- |Sort list by function result (don't use Schwarznegian transform!)
sortOn' f  =  sortBy (map2cmp f)

-- |Group list by function result
groupOn f  =  groupBy (map2eq f)

-- |Sort and Group list by function result
sort_and_groupOn  f  =  groupOn f . sortOn  f
sort_and_groupOn' f  =  groupOn f . sortOn' f

-- |,         
isAll f []       = True
isAll f [x]      = True
isAll f (x:y:ys) = f x y  &&  isAll f (y:ys)

-- |Check that list is sorted by given field/critery
isSortedOn f  =  isAll (<=) . map f

-- |Check that all elements in list are equal by given field/critery
isEqOn f      =  isAll (==) . map f

-- |Merge two lists, sorted by `cmp`, in one sorted list
merge :: (a -> a -> Ordering) -> [a] -> [a] -> [a]
merge cmp xs [] = xs
merge cmp [] ys = ys
merge cmp (x:xs) (y:ys)
 = case x `cmp` y of
        GT -> y : merge cmp (x:xs)   ys
        _  -> x : merge cmp    xs (y:ys)

-- |   `numGroups`     ,  `crit_f`
partitionList numGroups crit_f list =
  elems $ accumArray (flip (:)) [] (0, numGroups-1) (map (keyval crit_f) (reverse list))

-- partitionList numGroups crit_f list =
--   let xs = map (keyval crit_f) list
--       go 0 [] all = all
--       go n list prev = let (this, next) = partition (\(a,b) -> a==n-1) list
--                        in go (n-1) next (map snd this:prev)
--   in go numGroups xs []

-- |          `groups`:
--   splitList [(=='a'), (=='c')] 2 "cassa"  ->  ["aa","c","ss"]
--
splitList groups default_group filelist =
  let go [] filelist sorted  =  replaceAt default_group (reverse sorted) filelist
      go (group:groups) filelist sorted =
        let (found, notfound)  =  partition group filelist
        in go groups notfound (found:sorted)
  in go groups filelist []

-- |      `groups`,    `value`
findGroup groups default_group value  =  (findIndex ($ value) groups) `defaultVal` default_group

-- Utility functions for list operations
keyval  f x    =  (f x, x)                -- |Return pair containing computed key and original value
map2cmp f x y  =  (f x) `compare` (f y)   -- |Converts "key_func" to "compare_func"
map2eq  f x y  =  (f x) == (f y)          -- |Converts "key_func" to "eq_func"


-- |  
recursive f list  =  list &&& (x:recursive f xs)   where (x,xs) = f list

-- |   ,      `len_f`   
splitByLen len_f  =  recursive (\xs -> splitAt (len_f xs) xs)

-- |        `xs`    
splitByLens (len:lens) list  =  list &&& (x:splitByLens lens xs)    where (x,xs) = splitAt len list

-- |    ,   ,
--  "groupLen (fiSize) (+) (<16*mb) files"     ,
--       16 
groupLen mapper combinator tester  =  length . takeWhile tester . scanl1 combinator . map mapper

-- |  span  break: spanBreak isDigit "100a10b2c" = ("100a", "10b2c")
spanBreak crit xs  = let (s1,tail1) = span  crit xs
                         (s2,tail2) = break crit tail1
                     in (s1++s2, tail2)

{-# NOINLINE partitionList #-}
{-# NOINLINE splitList #-}


---------------------------------------------------------------------------------------------------
----    -------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------

listArray0 list  =  listArray (0,length(list)-1) list


---------------------------------------------------------------------------------------------------
----   tuples ----------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------

--   tuple/2
mapFst f (a,b)  =  (f a,   b)
mapSnd f (a,b)  =  (  a, f b)
map2   (f,g) a  =  (f a, g a)

--   tuple/3
fst3 (a,_,_)    =  a
snd3 (_,a,_)    =  a
thd3 (_,_,a)    =  a
map3 (f,g,h) a  =  (f a, g a, h a)


---------------------------------------------------------------------------------------------------
----    ------------------------------------------------------------------
---------------------------------------------------------------------------------------------------

infixl 0 =:, +=, -=, .=, .<-, <<=

-- Simple variables
ref = newIORef
val = readIORef
a=:b = writeIORef a b
a+=b = modifyIORef a (\a->a+b)
a-=b = modifyIORef a (\a->a-b)
a.=b = modifyIORef a (\a->b a)
a.<-b = modifyIORefIO a (\a->b a)
withRef init  =  with' (ref init) val


-- Accumulation lists
newtype AccList a = AccList [a]
newList   = ref$ AccList []
a<<=b     = a .= (\(AccList x) -> AccList$ b:x)
listVal a = val a >>== (\(AccList x) -> reverse x)
withList  =  with' newList listVal


-- |   ,    IORef
addToIORef :: IORef [a] -> a -> IO ()
addToIORef var x  =  var .= (x:)

-- | ,    IORef,  ,
--       ,   
modifyIORefIO :: IORef a -> (a -> IO a) -> IO ()
modifyIORefIO var action = do
  readIORef var  >>=  action  >>=  writeIORef var

-- |    
with' init finish action  =  do a <- init;  action a;  finish a


---------------------------------------------------------------------------------------------------
----       Int'    Integer' -----------
---------------------------------------------------------------------------------------------------

infixl 6 +:, -:
ptr+:n   = ptr `plusPtr` (fromIntegral n)
ptr-:buf = fromIntegral  (ptr `minusPtr` buf)
copyBytesI dst src len  =  copyBytes dst src (i len)
minI a b                =  i$ min (i a) (i b)
clipToMaxInt            =  minI (maxBound::Int)
atLeast                 =  max
i                       =  fromIntegral


---------------------------------------------------------------------------------------------------
----      ----------------------------------------------------
---------------------------------------------------------------------------------------------------

-- |        
--   heapsize     -  
--   aBUFFER_SIZE -    
--   aALIGN       -      ,   
--   returnBlock  -   ,  . ,
--                           
--
allocator heapsize aBUFFER_SIZE aALIGN returnBlock = do
  let aHEAP_START = 0          --  ,   = 0
      aHEAP_END   = heapsize   --  

  start <- ref aHEAP_START     --      
  end   <- ref aHEAP_END       --    
                               --    ,    
#if 0
  let debug = putStr         --  

  let printStats s = do      --     
        astart <- val start
        aend <- val end
        debug$ left_justify 48 s++"STATE start:"++show astart++", end:"++show aend++", avail:"++show ((aend-astart) `mod` aHEAP_END)++"\n"

  debug "\n"
#else
  let debug      = return
      printStats = return
#endif

  --     ,  aALIGN
  let align n  =  (((n-1) `div` aALIGN) + 1) * aALIGN

  --   : >=n,   aALIGN     aBUFFER_SIZE    
  let nextAvail n = if (aHEAP_END-aligned<aBUFFER_SIZE)
                      then aHEAP_END
                      else aligned
                    where aligned = align n

  --      
  let available = do
        astart <- val start
        aend   <- val end
        if (astart<=aend) then
           return (aend-astart)
         else if (astart<aHEAP_END) then
           return (aHEAP_END-astart)
         else do
           --        
           start =: aHEAP_START
           debug "===================================\n"
           printStats ""
           available

  --        
  let waitReleasingMemory = do
        (addr,size) <- returnBlock
        astart <- val start
        aend   <- val end
        unless (addr == aend || (addr==aHEAP_START && (aHEAP_END-aend<aBUFFER_SIZE)))$  fail "addToAvail!"
        let new_end = nextAvail(addr+size)
        if new_end == astart
          then do start=:aHEAP_START; end=:aHEAP_END -- now all memory is free
          else end =: new_end
        printStats$ "*** returned buf:"++show addr++" size:"++show size++"   "

  --     aBUFFER_SIZE.     - 
  --        
  let getBlock = do
        avail <- available
        if (avail >= aBUFFER_SIZE) then do
           block <- val start
           start =: error "Block not shrinked"
           return block
         else do
           waitReleasingMemory
           getBlock

  --      `size`.      getBlock
  let shrinkBlock block size = do
        astart <- val start
        --unless (astart == block)$      fail "Tryed to shrink another block"
        unless (size <= aBUFFER_SIZE)$  fail "Growing instead of shrinking :)"
        start =: nextAvail(block+size)
        printStats$ "getBlock buf:"++show block++", size: "++show aBUFFER_SIZE++" --> "++show size++"    "

  --      
  return (getBlock, shrinkBlock)


-- | ,    `heap`.
--  ,     `allocator`
memoryAllocator heap size chunksize align returnBlock = do
  let returnBlock2            =  do (buf,len) <- returnBlock; return (buf-:heap, len)
  (getBlock2, shrinkBlock2)  <-  allocator size chunksize align returnBlock2
  let getBlock                =  do block <- getBlock2; return (heap+:block)
      shrinkBlock buf len     =  do shrinkBlock2 (buf-:heap) len
  return (getBlock, shrinkBlock)



---------------------------------------------------------------------------------------------------
----    ---------------------------------------------------------------
---------------------------------------------------------------------------------------------------

-- |                               
data RegExpr = RE_End                     --                              ""
             | RE_Anything                --                             "*"
             | RE_FromEnd RegExpr         --   RE   '*':"bc"
             | RE_AnyChar RegExpr         --  ,  RE                 '?':"bc"
             | RE_Char    Char RegExpr    --  ,  RE              'a':"bc"
             | RE_FullRE  Regex           --         "r[0-9][0-9]"

-- |,      ,
--      
is_wildcard s  =  s `contains_one_of` "?*["

-- | RE     POSIX RE
translate_RE re = "^"++ (replaceAll "*" ".*"
                        .replaceAll "?" "."
                        .replaceAll "$" "\\$"
                        .replaceAll "[[[" "[^"   --   (       :)
                        .replaceAll "^" "\\^"
                        .replaceAll "[^" "[[["   -- 
                        .replaceAll "+" "\\+"
                        .replaceAll "." "\\.") re ++"$"

-- |       RegExpr
compile_RE s  =  case s of
  ""     -> RE_End
  "*"    -> RE_Anything
  '*':cs -> if ('*' `elem` cs) || ('[' `elem` cs)
              then RE_FullRE  (mkRegex$ translate_RE$ s)
              else RE_FromEnd (compile_RE$ reverse$   s)
  '[':cs -> RE_FullRE   (mkRegex$ translate_RE$ s)
  '?':cs -> RE_AnyChar  (compile_RE cs)
  c  :cs -> RE_Char   c (compile_RE cs)

-- |     
match_RE re s  =  case re of
  RE_End        -> null s
  RE_Anything   -> True
  RE_FullRE   r -> isJust (matchRegex r s)
  RE_FromEnd  r -> match_RE r (reverse s)
  RE_AnyChar  r -> case s of
                     ""   -> False
                     _:xs -> match_RE r xs
  RE_Char   c r -> case s of
                     ""   -> False
                     x:xs -> x==c && match_RE r xs

-- |   `s`   `re`
match re {-s-}  =  match_RE (compile_RE re) {-s-}

{-# NOINLINE translate_RE #-}

